from __future__ import annotations

import sys

import h5py
import numpy as np

from scitbx.array_family import flex
from serialtbx.detector.jungfrau import pad_stacked_format

from dxtbx.format.FormatNexus import FormatNexus
from dxtbx.format.nexus import h5str

"""
This class can read a Jungfrau image file that includes, in addition to the corrected data,
the gain maps, and the dark (pedestal) averages and RMS, providing the user access
to that data.
This was written to process SWISSFEL image files created using the CCTBX script jf16m_cxigeom2nexus.py

  libtbx.python cctbx_project/xfel/swissfel/jf16m_cxigeom2nexus.py \
  unassembled_file=X \
  raw_file=Y \
  geom_file=Z \
  pedestal_file=P \
  gain_file=G \
  detector_distance=D \
  include_spectra=True \
  output_file=O \
  raw=False \
  beam_file=B

where X,Y are the corrected,raw hdf5 files obtained at SWISSFEL for a particular run, Z is the crystFEL geom file
provided by SWISSFEL for the jungfrau, P is the hdf5 file provided by swissfel containing the pedestal data,  G is
the gainmap hdf5 file provided by SWISSFEL, D is the detector distance in mm, O is an output filename (to be read by
this format class) and B is the hdf5 file provided by SWISSFEL containing the wavelength spectra measurements (optional)
"""


class FormatNexusJungfrauExt(FormatNexus):
    @staticmethod
    def understand(image_file):
        if "xfel" not in sys.modules:
            return False
        with h5py.File(image_file, "r") as handle:
            # Note, in the future this class might support more instruments,
            # for now it is tested with the SwissFEL one

            if "name" not in handle["/entry/instrument"]:
                return False

            if (
                h5str(handle["/entry/instrument/name"][()])
                != "SwissFEL ARAMIS BEAMLINE ESB"
            ):
                return False

            try:
                data = handle["/entry/data"]
                if "pedestal" in data and "gains" in data and "raw" in data:
                    return True
            except (KeyError, AttributeError):
                pass
        return False

    def _slice_correction_array(self, array):
        """
        :param array: numpy array where the slow axis is image number
        and e.g. array[i] corresponds to the ith nexus data entry
        :return: array sliced according to the nexus geometry specification
        """
        return np.array(
            [
                array[:, self._slices[i][0], self._slices[i][1]]
                for i in range(len(self._slices))
            ]
        )

    def _start(self):
        super()._start()
        data_handle = self._reader.entries[0].data[0].handle

        # nexus slices
        self._slices = self._raw_data._datalists[0]._all_slices

        self._gain = self._slice_correction_array(data_handle["gains"][()])
        self._pedestal = self._slice_correction_array(data_handle["pedestal"][()])
        self._pedestalRMS = self._slice_correction_array(data_handle["pedestalRMS"][()])

        self.dxtbx_detector_shape = self._gain[:, 0].shape

    def get_14bit_component(self, index, as_flex=False):
        """
        Jungfrau raw data are stored in 16-bit numbers, the 2 significant bits
        encode the gain mode (a number 0,1,2,3), and the other bits store the measurement
        (sum of a photon scattering term, and a pedestal term)
        This function returns the latter term (14-bit measurement)
        :param index: int, shot index
        :param as_flex: bool, return as tuple of flex arrays
        """
        data_handle = self._reader.entries[0].data[0].handle
        raw_16bit = data_handle["raw"][index]
        raw_14bit = pad_stacked_format(raw_16bit & 0x3FFF)
        raw_14bit = self._slice_correction_array(np.array([raw_14bit]))[:, 0]
        if as_flex:
            raw_14bit = tuple([flex.double(d.astype(np.float64)) for d in raw_14bit])
        return raw_14bit

    def get_pedestal_rms(
        self, index, as_flex=False, return_gain_modes=False, divide_by_gain=True
    ):
        """
        This method uses the dynamic (per-shot) pixel gain mode to lookup the per-shot pedestal RMS
        This is important for detailed error analysis of pixel measurements.
        :param index: image index (0-num_images)
        :param as_flex: if True, return tuple of flex arrays (like get_raw_data), else return as a numpy array
        :param return_gain_modes: whether to return the gain modes in addition to the pedestalRMS
        :param divide_by_gain: normalize the pedestalRMS such that the units are the same as get_raw_data (keV)
        """
        if self._pedestalRMS is None:
            raise AttributeError("There is not pedestalRMS in the master file ... ")
        data_handle = self._reader.entries[0].data[0].handle
        raw = data_handle["raw"][index]
        shot_gain_modes = pad_stacked_format(raw >> 14)
        shot_gain_modes = self._slice_correction_array(np.array([shot_gain_modes]))
        shot_gain_modes = shot_gain_modes[:, 0]
        shot_rms = np.zeros(self.dxtbx_detector_shape, np.float64)
        for mode in 0, 1, 3:
            gain_index = mode if mode != 3 else 2  # jungfrau quirk
            sel = shot_gain_modes == mode  # pixels in the gain mode
            pedestalRms = self._pedestalRMS[:, gain_index]
            if divide_by_gain:
                gain = self._gain[:, gain_index]
                shot_rms[sel] = pedestalRms[sel] / np.abs(gain[sel])
            else:
                shot_rms[sel] = pedestalRms[sel]
        if as_flex:
            shot_rms = tuple([flex.double(p) for p in shot_rms])
        return_packet = shot_rms
        if return_gain_modes:
            return_packet = shot_rms, shot_gain_modes
        return return_packet



from __future__ import annotations

import os
import sys

from iotbx.detectors.adsc import ADSCImage

from dxtbx.format.FormatSMVADSC import FormatSMVADSC


class FormatSMVADSCNoDateStamp(FormatSMVADSC):
    """A class for reading SMV format ADSC images, with detector serial number"""

    @staticmethod
    def understand(image_file):
        # assert for this that the image file has to be a file not a URL
        if not os.path.exists(image_file):
            return False

        size, header = FormatSMVADSC.get_smv_header(image_file)

        wanted_header_items = ["TIME"]
        if any(item not in header for item in wanted_header_items):
            return False

        unwanted_header_items = ["DATE"]

        if any(item in header for item in unwanted_header_items):
            return False

        return True

    def _scan(self):
        """Return the scan information for this image, using the timestamp
        from the file rather than the header."""

        exposure_time = float(self._header_dictionary["TIME"])

        epoch = float(os.stat(self._image_file)[8])

        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def detectorbase_start(self):
        self.detectorbase = ADSCImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCNoDateStamp.understand(arg))



from __future__ import annotations

import argparse

import h5py

import dxtbx.util

sample = None


def depends_on(in_name):
    f = h5py.File(in_name, "r")

    depends = {}

    global sample

    def finder(thing, path):
        if hasattr(thing, "attrs"):
            if "depends_on" in thing.attrs:
                depends[path] = thing.attrs["depends_on"]
            if thing.attrs.get("NX_class", None) == "NXsample":
                global sample
                sample = path
        if hasattr(thing, "keys"):
            for k in thing:
                try:
                    finder(thing[k], path=f"{path}/{k}")
                except (OSError, TypeError, ValueError, KeyError):
                    pass

    # clean up hierarchy to just have sample stuff

    finder(f, path="")
    delete = []

    for d in sorted(depends):
        if "/entry/sample/transformations" not in d:
            delete.append(d)

    for d in delete:
        del depends[d]

    # invert for printing

    inverted = {}

    for d in depends:
        t = depends[d]
        if t in inverted:
            print(t, inverted[t], d)
        inverted[t] = d

    print("Dependency hierarchy in file:")
    at = "."
    depth = 0
    while at in inverted:
        print(f"{' ' * depth}+ {at}")
        at = inverted[at]
        depth += 2
    print(f"{' ' * depth}+ {at}")
    print("")

    print(f"Sample at {sample} depends on:")
    if "depends_on" in f[sample]:
        print(f[sample]["depends_on"][()])
    elif hasattr(f[sample], "attrs") and "depends_on" in f[sample].attrs:
        print(f[sample].attrs["depends_on"])
    else:
        print(f"{sample} -> depends_on not found")

    f.close()


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(
        description="Print depends_on hierarchy for Nexus files"
    )
    parser.add_argument("filename", help="The nexus file")
    opts = parser.parse_args(args)
    depends_on(opts.filename)


if __name__ == "__main__":
    run()



from __future__ import annotations

import glob
import os
import uuid

from dxtbx.serialize.filename import resolve_path


def test_resolve_path(monkeypatch):
    # Resolve path
    def alwaystrue(path):
        return True

    # Set an environment variable
    monkeypatch.setenv("HELLO_WORLD", "EXPANDED")

    # First try a path that does not exist
    filename = str(uuid.uuid4())
    new_path = os.path.join("~", "$HELLO_WORLD", filename)
    path = resolve_path(new_path)
    assert path == new_path

    # Now pretend the path exists by ensuring its glob is truthy
    monkeypatch.setattr(glob, "glob", alwaystrue)
    path = resolve_path(new_path)
    assert path == os.path.join(os.path.expanduser("~"), "EXPANDED", filename)

    new_path = os.path.join("$HELLO_WORLD", filename)
    path = resolve_path(new_path)
    assert path == os.path.abspath(os.path.join("EXPANDED", filename))



from __future__ import annotations

import os.path
import shutil
import subprocess


def test_plot_models_noninteractive_to_pdf(dials_data, tmp_path):
    ssx = dials_data("cunir_serial_processed")
    expts = str(ssx / "imported_no_ref_5.expt")
    subprocess.run(
        [shutil.which("dxtbx.plot_detector_models"), expts, "pdf_file=plot_test.pdf"],
        cwd=tmp_path,
        capture_output=True,
    )
    assert os.path.exists(tmp_path / "plot_test.pdf")



#
# See https://github.com/dials/dials/wiki/pytest for documentation on how to
# write and run pytest tests, and an overview of the available features.
#

from __future__ import annotations

import pytest

collect_ignore = []


def pytest_addoption(parser):
    """Add '--regression' options to pytest."""
    try:
        parser.addoption(
            "--regression",
            action="store_true",
            default=False,
            help="run (time-intensive) regression tests",
        )
    except ValueError:
        # Thrown in case the command line option is already defined
        pass


def pytest_collection_modifyitems(config, items):
    """Tests marked as regression are only run with --regression."""
    if not config.getoption("--regression"):
        skip_regression = pytest.mark.skip(reason="Test only runs with --regression")
        for item in items:
            if "regression" in item.keywords:
                item.add_marker(skip_regression)



from __future__ import annotations

import math
import random

import pytest

from cctbx import factor_ev_angstrom
from scitbx.array_family import flex

from dxtbx.model import Spectrum


def test_spectrum():
    mean_ev_1 = factor_ev_angstrom  # 1.0Å
    wavelengths = 2 / 3, 1, 3 / 2

    for wavelength in wavelengths:
        mean_ev = mean_ev_1 / wavelength
        spectrum = Spectrum(
            flex.double([mean_ev - 50, mean_ev, mean_ev + 50]),
            flex.double([0.5, 1.0, 0.5]),
        )
        assert spectrum.get_weighted_energy_eV() == pytest.approx(mean_ev)
        assert spectrum.get_weighted_wavelength() == pytest.approx(wavelength)

        spectrum = Spectrum.from_dict(spectrum.to_dict())
        assert spectrum.get_weighted_energy_eV() == pytest.approx(mean_ev)
        assert spectrum.get_weighted_wavelength() == pytest.approx(wavelength)


def test_spectrum_bandwidth():
    e, c = __generate_spectrum()

    spectrum = Spectrum(e, c)

    emin, emax = spectrum.get_emin_eV(), spectrum.get_emax_eV()

    assert emin == pytest.approx(11840, abs=2)
    assert emax == pytest.approx(11876, abs=2)


def test_spectrum_weighted_mean_variance():
    p = 11900.0
    h = 10000.0
    w = 5.0

    e, c = __generate_simple(p, h, w)

    spectrum = Spectrum(e, c)

    mean = spectrum.get_weighted_energy_eV()
    variance = spectrum.get_weighted_energy_variance()

    assert mean == pytest.approx(p, abs=1e-3)
    assert variance == pytest.approx(w * w, abs=1e-3)


def __generate_spectrum():
    energies = flex.double()
    counts = flex.double()

    peaks, heights, widths = (
        (11860, 11840, 11853, 11860, 11865, 11870),
        (10000, 5000, 5000, 5000, 5000),
        (10, 1, 1, 1, 1),
    )

    for dev in range(118000, 120000):
        ev = 0.1 * dev
        c = random.random() * 100 - 50
        for p, h, w in zip(peaks, heights, widths):
            c += h * math.exp(-(((ev - p) / w) ** 2))
        energies.append(ev)
        counts.append(c)

    return energies, counts


def __generate_simple(peak, height, width):
    energies = flex.double()
    counts = flex.double()

    for dev in range(118000, 120000):
        ev = 0.1 * dev
        c = height * math.exp(-0.5 * (((ev - peak) / width) ** 2))
        energies.append(ev)
        counts.append(c)

    return energies, counts



from __future__ import annotations

import xml
import xml.etree.ElementTree as ET
from multiprocessing import Pool, cpu_count
from os.path import join
from sys import argv
from typing import Optional, Tuple

import h5py
import numpy as np

from cctbx.array_family import flex

import dxtbx_flumpy as flumpy
from dxtbx import IncorrectFormatError
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.model import Detector
from dxtbx.model.beam import BeamFactory, PolychromaticBeam, Probe
from dxtbx.model.goniometer import Goniometer, GoniometerFactory
from dxtbx.model.scan import Scan, ScanFactory
from dxtbx.model.tof_helpers import InstrumentDefinitionReader


class FormatMANDI(FormatHDF5):
    """
    Class to read NXTOFRAW from MaNDi
    (https://neutrons.ornl.gov/mandi)
    """

    def __init__(self, image_file: str, **kwargs) -> None:
        super().__init__(image_file)
        if not FormatMANDI.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self.image_file = image_file
        self.nxs_file = self.open_file(image_file)
        self._base_entry = self.get_base_entry_name(self.nxs_file)
        self.detector = None
        self.raw_data = None
        self.xml_reader = InstrumentDefinitionReader()
        self.xml_file = self.get_xml_file()

    def open_file(self, image_file_path: str) -> h5py.File:
        return h5py.File(image_file_path, "r")

    def get_base_entry_name(self, nxs_file: h5py.File) -> str:
        return list(nxs_file.keys())[0]

    def get_xml_file(self) -> xml.etree.ElementTree.Element:
        xml_string = self.nxs_file[self._base_entry]["instrument/instrument_xml/data"][
            0
        ].decode()
        return ET.fromstring(xml_string)

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatMANDI.is_mandi_file(image_file)
        except OSError:
            return False

    @staticmethod
    def is_mandi_file(image_file: str) -> bool:
        def get_name(image_file: str) -> str:
            with h5py.File(image_file, "r") as handle:
                if len(handle) == 0:
                    return ""
                base_entry = list(handle.keys())[0]
                if f"{base_entry}/instrument/name" not in handle:
                    return ""
                try:
                    return handle[f"{base_entry}/instrument/name"][0].decode()
                except (ValueError, IndexError):
                    return ""

        return get_name(image_file) == "MANDI"

    def get_instrument_name(self) -> str:
        return self.nxs_file[self._base_entry]["instrument/name"][0].decode()

    def get_experiment_title(self) -> str:
        return self.nxs_file[self._base_entry]["title"][0].decode()

    def get_experiment_run_number(self) -> str:
        return str(self.nxs_file[self._base_entry]["run_number"][0].decode())

    def get_experiment_description(self) -> str:
        title = self.get_experiment_title()
        run_number = self.get_experiment_run_number()
        return f"{title} ({run_number})"

    def get_raw_data(self, index: int) -> Tuple[flex.int]:
        raw_data = []
        panel_size = self._get_image_size()
        for panel_name in self._get_panel_names():
            spectra = np.reshape(
                self.nxs_file[self._base_entry][f"{panel_name}_events"]["spectra"][
                    :, index : index + 1
                ],
                panel_size,
            ).T
            raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        return tuple(raw_data)

    def get_detector(self, index: int = None) -> Detector:
        num_panels = self._get_num_panels()
        panel_names = self._get_panel_names()
        panel_type = self._get_panel_type()
        image_size = self._get_image_size()
        trusted_range = self._get_panel_trusted_range()
        pixel_size = self._get_pixel_size()
        panel_origins, fast_axes, slow_axes = (
            self.xml_reader.get_dials_detector_geometry(self.xml_file)
        )
        gain = self._get_panel_gain()
        panel_projections = self._get_panel_projections_2d()
        detector = Detector()
        root = detector.hierarchy()

        for i in range(num_panels):
            group = root.add_group()
            panel = group.add_panel()
            panel.set_type(panel_type)
            panel.set_name(panel_names[i])
            panel.set_image_size(image_size)
            panel.set_trusted_range(trusted_range)
            panel.set_pixel_size(pixel_size)
            panel.set_local_frame(fast_axes[i], slow_axes[i], panel_origins[i])
            panel.set_gain(gain)
            r, t = panel_projections[i + 1]
            r = tuple(map(int, r))
            t = tuple(map(int, t))
            panel.set_projection_2d(r, t)

        return detector

    def _get_time_channel_bins(self) -> list[float]:
        # (usec)
        return self.nxs_file[self._base_entry]["time_of_flight"][:]

    def _get_time_of_flight(self) -> list[float]:
        # (usec)
        bins = self._get_time_channel_bins()
        return tuple([float(bins[i] + bins[i + 1]) * 0.5 for i in range(len(bins) - 1)])

    def _get_sample_to_source_distance(self) -> float:
        # (mm)
        return 30000

    def _get_sample_to_source_direction(self) -> tuple[float, float, float]:
        return (0, 0, -1)

    def _get_num_panels(self) -> int:
        return self.xml_reader.get_num_panels(self.xml_file)

    def _get_panel_names(self) -> tuple[str]:
        return self.xml_reader.get_panel_names(self.xml_file)

    def _get_panel_gain(self) -> float:
        return 1.0

    def _get_panel_trusted_range(self) -> tuple[int, int]:
        return (-1, 100000)

    def get_num_images(self) -> int:
        return len(self._get_time_of_flight())

    def get_beam(self, index: int = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = (2.0, 4.0)
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            wavelength_range=wavelength_range,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
        )

    def get_scan(self, index=None) -> Scan:
        image_range = (1, self.get_num_images())
        properties = {"time_of_flight": self._get_time_of_flight()}
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties=properties
        )

    def get_goniometer(self, idx: int = None) -> Goniometer:
        rotation_axis_phi = (0.0, 1.0, 0.0)
        rotation_axis_omega = (0.0, 1.0, 0.0)
        rotation_axis_chi = (0.0, 0.0, 1.0)
        fixed_rotation = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)

        goniometer = GoniometerFactory.make_goniometer(
            rotation_axis_phi, fixed_rotation
        )
        phi = self.nxs_file["entry/DASlogs/phi/average_value"][0]
        omega = self.nxs_file["entry/DASlogs/omega/average_value"][0]
        chi = self.nxs_file["entry/DASlogs/chi/average_value"][0]

        goniometer.rotate_around_origin(rotation_axis_phi, phi)
        goniometer.rotate_around_origin(rotation_axis_chi, chi)
        goniometer.rotate_around_origin(rotation_axis_omega, omega)
        return goniometer

    def _get_image_size(self) -> tuple[int, int]:
        return (256, 256)

    def _get_pixel_size(self) -> tuple[float, float]:
        return (0.618, 0.618)

    def _get_panel_type(self) -> str:
        return "SENSOR_PAD"

    def _get_panel_projections_2d(self) -> dict:
        p_w, p_h = self._get_image_size()
        panel_pos = {}
        count = 1
        for i in range(8):
            for j in range(5):
                panel_pos[count] = ((1, 0, 0, 1), (p_h * i, p_w * j))
                count += 1
        panel_pos[count] = ((1, 0, 0, 1), (p_h * i, p_w * j))
        return panel_pos

    @staticmethod
    def add_histogram_data_to_nxs_file(
        nxs_file_path: str,
        remove_event_data: bool,
        spectra_output_name: str = "spectra",
        write_tof_bins: bool = True,
        delta_tof: float = 5,  # (usec)
        tof_padding: float = 100,  # (usec)
        panel_size: tuple[int, int] = (256, 256),  # (px)
        nproc: int = 8,
    ) -> None:
        tof_bins = FormatMANDI.generate_tof_bins(
            nxs_file=nxs_file_path,
            delta_tof=delta_tof,
            padding=tof_padding,
        )
        FormatMANDI.write_histogram_data(
            nxs_file_path=nxs_file_path,
            tof_bins=tof_bins,
            panel_size=panel_size,
            remove_event_data=remove_event_data,
            spectra_output_name=spectra_output_name,
            write_tof_bins=write_tof_bins,
        )

    @staticmethod
    def write_histogram_data(
        nxs_file_path: str,
        tof_bins: np.array,
        panel_size: tuple[int, int],
        remove_event_data: bool,
        spectra_output_name: str = "spectra",
        write_tof_bins: bool = True,
        nproc: int = 8,
    ) -> None:
        """
        Generates histogram spectra for a given detector and writes it to nxs_file
        """

        def delete_event_data(nxs_file, base_dir, panel_name):
            del nxs_file[join(join(base_dir, panel_name), "event_index")]
            del nxs_file[join(join(base_dir, panel_name), "event_id")]
            del nxs_file[join(join(base_dir, panel_name), "event_time_zero")]
            del nxs_file[join(join(base_dir, panel_name), "event_time_offset")]

        print(f"Writing histrogram data in {nxs_file_path}")
        print(f"Remove event data: {remove_event_data}")
        nxs_file = h5py.File(nxs_file_path, "r+")
        base_dir = list(nxs_file.keys())[0]

        panel_names = FormatMANDI.get_panel_names(nxs_file)
        written_tof_bins = False
        for panel_name in panel_names:
            print(f"Processing panel {panel_name}")
            output_path = join(base_dir, panel_name)
            output_path = join(output_path, spectra_output_name)
            print(f"Writing spectra to {output_path}")
            panel_spectra = FormatMANDI.generate_histogram_data_for_panel(
                nxs_file, tof_bins, panel_size, panel_name
            )
            nxs_file.create_dataset(output_path, data=panel_spectra, compression="gzip")
            if remove_event_data:
                delete_event_data(nxs_file, base_dir, panel_name)
                print(f"Removed event data for {panel_name}")
            if write_tof_bins and not written_tof_bins:
                tof_path = join(base_dir, "time_of_flight")
                print(f"Writing time of flight bins to {tof_path}")
                nxs_file.create_dataset(tof_path, data=tof_bins, compression="gzip")
                written_tof_bins = True

        nxs_file.close()

    @staticmethod
    def compute_event_histogram(
        args: tuple[int, np.array, np.array, np.array],
    ) -> np.array:
        pixel_idx, event_time_offset, corrected_event_id, tof_bins = args
        h, _ = np.histogram(
            event_time_offset[corrected_event_id == pixel_idx], tof_bins
        )
        return h

    @staticmethod
    def generate_histogram_data_for_panel(
        nxs_file: h5py.File,
        tof_bins: np.array,
        panel_size: tuple[int, int],
        panel_name: str,
        nproc=8,
    ) -> np.array:
        """
        Generates histogram data for a given panel
        """

        ## Get panel data
        panel_number = FormatMANDI.get_panel_number(panel_name)
        # Actual pixel ids, starting from bottom left and going up y axis first
        event_id = nxs_file[f"entry/{panel_name}/event_id"][:]
        # Time each event_id was triggered after event_time_zero (sec)
        event_time_offset = nxs_file[f"entry/{panel_name}/event_time_offset"][:]

        # event_ids are given with an offset
        event_id_offset = panel_number * panel_size[0] * panel_size[1]
        corrected_event_id = event_id - event_id_offset

        num_pixels = panel_size[0] * panel_size[1]
        spectra = np.zeros((num_pixels, len(tof_bins) - 1), dtype=np.int32)

        num_cpu = cpu_count()
        if nproc > num_cpu:
            nproc = num_cpu

        pool = Pool(processes=nproc)

        args_list = [
            (i, event_time_offset, corrected_event_id, tof_bins)
            for i in range(num_pixels)
        ]
        results = pool.map(FormatMANDI.compute_event_histogram, args_list)

        pool.close()
        pool.join()

        for i, h in enumerate(results):
            spectra[i] = h

        return spectra

    @staticmethod
    def get_time_range_for_panel(
        nxs_file: h5py.File, panel_name: str
    ) -> Optional[Tuple[float, float]]:
        event_index = nxs_file[f"entry/{panel_name}/event_index"][:]
        event_time_zero = nxs_file[f"entry/{panel_name}/event_time_zero"][:]
        event_time_offset = nxs_file[f"entry/{panel_name}/event_time_offset"][:]

        tofs_min = []
        tofs_max = []

        n_pulses = len(event_index) - 1

        for p in range(n_pulses):
            start = event_index[p]
            stop = event_index[p + 1]

            if start == stop:
                continue

            T0 = event_time_zero[p]
            offsets = event_time_offset[start:stop]

            tofs_min.append(T0 + np.min(offsets))
            tofs_max.append(T0 + np.max(offsets))

        if not tofs_min:
            return None, None

        return min(tofs_min), max(tofs_max)

    @staticmethod
    def get_time_range_for_dataset(nxs_file_path: str) -> tuple[float, float]:
        """
        Iterates over num_panels to find the overall min/max tof event recorded
        """

        # TODO get panel_size and panel_number from nxs_file xml

        nxs_file = h5py.File(nxs_file_path, "r")

        min_tof = -1
        max_tof = -1

        panel_names = FormatMANDI.get_panel_names(nxs_file)

        for panel_name in panel_names:
            try:
                min_event_time, max_event_time = FormatMANDI.get_time_range_for_panel(
                    nxs_file, panel_name
                )
                if min_event_time is None or max_event_time is None:
                    # Some banks contain no data
                    continue
                if min_tof < 0 or min_event_time < min_tof:
                    min_tof = min_event_time
                if max_event_time > max_tof:
                    max_tof = max_event_time
            except KeyError:
                # Some banks not always present
                pass

        nxs_file.close()

        return min_tof, max_tof

    @staticmethod
    def generate_tof_bins(
        nxs_file: str,
        delta_tof: float = 50,
        padding: float = 100,
    ) -> np.ndarray:
        """
        delta_tof: float (usec)
        padding: float (usec)
        """

        min_tof, max_tof = FormatMANDI.get_time_range_for_dataset(nxs_file)
        min_tof = min_tof - padding
        max_tof = max_tof + padding
        print(
            f"Time of flight range for {nxs_file}: {round(min_tof, 3)} - {round(max_tof, 3)} (usec)"
        )
        num_bins = int((max_tof - min_tof) / delta_tof)
        return np.linspace(min_tof, max_tof, num_bins)

    @staticmethod
    def get_panel_names(nxs_file: h5py.File) -> list[str]:
        raw_names = [i for i in nxs_file[list(nxs_file.keys())[0]] if "bank" in i]
        names = []
        for name in raw_names:
            try:
                entry = (name, int("".join([j for j in name if j.isdigit()])))
                names.append(entry)
            except ValueError:  # Other fields with bank in them
                continue
        return [i[0] for i in sorted(names, key=lambda x: x[0])]

    @staticmethod
    def get_panel_number(panel_name: str) -> int:
        return int("".join([i for i in panel_name if i.isdigit()]))


if __name__ == "__main__":
    for arg in argv[1:]:
        print(FormatMANDI.understand(arg))



from __future__ import annotations

import os

import pytest

from dxtbx.model.scan import ScanFactory
from dxtbx.model.scan_helpers import scan_helper_image_files


@pytest.fixture
def image_test_dir(tmp_path):
    """Create a sequence of 20 'images' for scan tests to pick up"""
    for i in range(1, 21):
        image_file = tmp_path / f"image_{i:03d}.dat"
        image_file.touch()
    return str(tmp_path)


def test_helper_image_files(image_test_dir):
    """Test the static methods in scan_helper_image_files."""
    assert scan_helper_image_files()

    template = "image_###.dat"

    assert (
        len(
            scan_helper_image_files.template_directory_to_indices(
                template, image_test_dir
            )
        )
        == 20
    )

    assert scan_helper_image_files.template_directory_index_to_image(
        template, image_test_dir, 1
    ) == os.path.join(image_test_dir, "image_001.dat")

    assert (
        scan_helper_image_files.template_index_to_image(template, 1) == "image_001.dat"
    )

    assert scan_helper_image_files.image_to_template_directory(
        os.path.join(image_test_dir, "image_001.dat")
    ) == (template, image_test_dir)

    assert scan_helper_image_files.image_to_index("image_001.dat") == 1

    assert scan_helper_image_files.image_to_template("image_001.dat") == "image_###.dat"

    assert scan_helper_image_files.image_to_index("image_6.8kev_1_001.cbf") == 1


def test_xScanFactory(image_test_dir):
    """Test out the ScanFactory."""

    template = "image_###.dat"

    xscans = [
        ScanFactory.single_file(
            scan_helper_image_files.template_directory_index_to_image(
                template, image_test_dir, j + 1
            ),
            1.0,
            18 + 0.5 * j,
            0.5,
            j,
        )
        for j in range(20)
    ]

    xscans.reverse()

    with pytest.raises(RuntimeError):
        print(sum(xscans[1:], xscans[0]))

    xscans.sort()

    sum(xscans[1:], xscans[0])

    a = ScanFactory.add(xscans[:10])
    b = ScanFactory.add(xscans[10:])

    a + b

    filename = scan_helper_image_files.template_directory_index_to_image(
        template, image_test_dir, 1
    )

    assert len(ScanFactory.search(filename)) == 20

    (a + b)[1:5]
    (a + b)[:10]



from __future__ import annotations

import h5py

from dxtbx.format.FormatNexus import FormatNexus


class FormatNexusTimepix2M(FormatNexus):
    @staticmethod
    def understand(image_file):
        # Get the file handle
        with h5py.File(image_file, "r") as handle:
            if "/entry/instrument/detector/module/data_size" in handle:
                size = handle["/entry/instrument/detector/module/data_size"]
                if tuple(size[()]) == (1147, 2069):
                    return True

        return False

    def __init__(self, image_file, **kwargs):
        super().__init__(image_file, **kwargs)

    def _detector(self):
        """return a detector with additional masking"""

        detector = super()._detector()

        # add masks for intermediate regions:

        module_spacing = 117
        chip_spacing = 3

        mask_width = 5

        nfast = 8 * 256 + 7 * chip_spacing
        nslow = 2 * 256 + chip_spacing

        for module in 0, 1:
            # horizontal masks
            for j in range(1):
                nn = module * (nslow + module_spacing)
                detector[0].add_mask(0, nn + 254, nfast, nn + 254 + mask_width)
            # vertical masks
            for j in range(1, 8):
                nn = module * (nslow + module_spacing)
                mm = j * 259
                detector[0].add_mask(mm - 5, nn, mm, nn + nslow)

        # between modules
        detector[0].add_mask(0, nslow, nfast, nslow + module_spacing)

        return detector



from __future__ import annotations

import logging
from os.path import join

import h5py
import numpy as np

import cctbx.array_family.flex as flex

import dxtbx_flumpy as flumpy
from dxtbx import IncorrectFormatError
from dxtbx.format.FormatESSNMX import FormatESSNMX
from dxtbx.model.beam import BeamFactory, PolychromaticBeam, Probe
from dxtbx.model.scan import Scan, ScanFactory

logger = logging.getLogger(__name__)


class FormatESSNMXLaue(FormatESSNMX):
    """
    Class to read files from NMX
    https://europeanspallationsource.se/instruments/nmx
    flattened in the TOF dimension

    To add Laue data to a MANDI NeXus file:
    >>> from dxtbx.format.FormatMANDILaue import FormatMANDILaue
    >>> nxs_file = "/path/to/file/example.nxs.h5"
    >>> FormatESSNMXLaue.add_laue_data_to_nxs_file(nxs_file)

    """

    def __init__(self, image_file, **kwargs) -> None:
        if not FormatESSNMXLaue.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        self._nxs_file = h5py.File(image_file, "r")
        self._raw_data = None

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatESSNMXLaue.is_nmx_laue_file(image_file)
        except (OSError, KeyError):
            return False

    @staticmethod
    def is_nmx_laue_file(image_file: str) -> bool:
        def get_name(image_file):
            try:
                with h5py.File(image_file, "r") as handle:
                    return handle["entry/instrument/name"][...].item().decode()
            except (OSError, KeyError, AttributeError):
                return ""

        if get_name(image_file) == "NMX":
            try:
                with h5py.File(image_file, "r") as handle:
                    return "laue_data" in handle["entry/instrument/detector_panel_0"]

            except (ValueError, IndexError):
                return False

    def get_num_images(self) -> int:
        return 1

    def get_scan(self, index=None) -> Scan:
        image_range = (1, self.get_num_images())
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties={}
        )

    def get_beam(self, index: int | None = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = self._get_wavelength_range()
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
            wavelength_range=wavelength_range,
        )

    def get_raw_data(
        self,
        index: int,
    ) -> tuple[flex.int]:
        raw_data = []
        for panel in self._get_panels():
            spectra = panel["laue_data"][:]
            spectra = panel["laue_data"].astype(np.int32)
            raw_data.append(flumpy.from_numpy(np.ascontiguousarray(spectra)))

        return tuple(raw_data)

    @staticmethod
    def get_panel_names(nxs_file: h5py.File) -> list[str]:
        panel_names = []
        inst_dset = nxs_file["/entry/instrument/"]
        for name, dset in inst_dset.items():
            if dset.attrs.get("NX_class") == "NXdetector":
                panel_names.append(name)
        return panel_names

    @staticmethod
    def add_laue_data_to_nxs_file(
        nxs_file_path: str,
        output_name: str = "laue_data",
        compress=True,
    ) -> None:
        """
        Extracts event data and writes out laue data for each panel in
        nxs_file_path
        """

        print(f"Writing laue data in {nxs_file_path}")

        nxs_file = h5py.File(nxs_file_path, "r+")
        base_dir = "entry/instrument"

        panel_names = FormatESSNMXLaue.get_panel_names(nxs_file)
        for panel_name in panel_names:
            output_path = join(base_dir, panel_name)
            output_path = join(output_path, output_name)
            panel_data = FormatESSNMXLaue.generate_laue_data_for_panel(
                nxs_file, panel_name
            )
            if compress:
                nxs_file.create_dataset(
                    output_path, data=panel_data, compression="gzip"
                )
            else:
                nxs_file.create_dataset(output_path, data=panel_data)

        nxs_file.close()

    @staticmethod
    def generate_laue_data_for_panel(
        nxs_file: h5py.File,
        panel_name: str,
    ) -> np.array:
        """
        Generates laue data for a given panel
        """

        data = nxs_file[f"entry/instrument/{panel_name}/data"][:]

        laue_data = np.sum(data, axis=2)

        return laue_data



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for example on ALS beamline 8.2.2 from back in the
day which had its own way of recording beam centre.
"""

from __future__ import annotations

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN905(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 905."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 905."""

        # check this is detector serial number 905

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 905

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        if (
            "DENZO_XBEAM" in self._header_dictionary
        ):  # Beamline group changed key=value tags several times
            beam_x = float(self._header_dictionary["DENZO_XBEAM"])
            beam_y = float(self._header_dictionary["DENZO_YBEAM"])
        elif "DENZO_BEAM_CENTER_X" in self._header_dictionary:
            beam_x = float(self._header_dictionary["DENZO_BEAM_CENTER_X"])
            beam_y = float(self._header_dictionary["DENZO_BEAM_CENTER_Y"])
        else:
            raise Exception("cannot find beam center for detector S/N 905")
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        trusted_range = self._adsc_trusted_range()

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            trusted_range,
            [],
            gain=self._adsc_module_gain(),
            pedestal=40,
        )



from __future__ import annotations

import random

from scitbx import matrix

import dxtbx
from dxtbx.model import parallax_correction, parallax_correction_inv


def correct_gold(detector, attlen, xy):
    s1 = matrix.col(detector.get_lab_coord(xy))
    lab = attlen * s1 / s1.length()
    d = detector.get_d_matrix()
    d0 = matrix.col((d[0], d[3], d[6]))
    d1 = matrix.col((d[1], d[4], d[7]))
    mm0 = d0.dot(lab) / d0.length()
    mm1 = d1.dot(lab) / d1.length()
    mmcal = matrix.col((xy[0] + mm0, xy[1] + mm1))
    return mmcal


def test_run(dials_data):
    filename = str(dials_data("misc_regression") / "sim_mx-GXPARM.XDS")

    models = dxtbx.load(filename)
    detector = models.get_detector()
    assert len(detector) == 1
    detector = detector[0]
    attlen = 0.252500934883
    distance = detector.get_distance()
    origin = detector.get_ray_intersection(detector.get_normal())

    for i in range(10000):
        # Generate some random coordinates
        xy = matrix.col((random.uniform(-1000, 1000), random.uniform(-1000, 1000)))

        # Do the forward and reverse corrections
        corr_gold = matrix.col(correct_gold(detector, attlen, xy))
        corr = matrix.col(parallax_correction(distance, attlen, origin, xy))
        corr_inv = matrix.col(parallax_correction_inv(distance, attlen, origin, corr))

        # Check the values
        assert abs(corr_gold - corr) < 1e-7
        assert abs(corr_inv - xy) < 1e-3



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSCSN, customised for example on Australian Synchrotron SN 928
which has reversed phi.
"""

from __future__ import annotations

import sys

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN928(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 928."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 928."""

        # check this is detector serial number 928

        size, header = FormatSMVADSCSN.get_smv_header(image_file)
        if int(header["DETECTOR_SN"]) != 928:
            return False

        return True

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header."""

        return self._goniometer_factory.single_axis_reverse()

    def _scan(self):
        """Return the scan information for this image. There may be
        no timestamps in there..."""

        exposure_time = float(self._header_dictionary["TIME"])
        epoch = 0
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCSN928.understand(arg))
        print(FormatSMVADSCSN928(arg).get_scan())
        print(FormatSMVADSCSN928(arg).get_goniometer())
        print(FormatSMVADSCSN928(arg).get_detector())
        print(FormatSMVADSCSN928(arg).get_beam())



"""Methods specific to interacting with CSPAD images"""

from __future__ import annotations

import pycbf

from cctbx.eltbx import attenuation_coefficient
from scitbx.matrix import col, sqr

from dxtbx.format.FormatCBFFull import FormatCBFFullStill
from dxtbx.format.FormatCBFMultiTileHierarchy import FormatCBFMultiTileHierarchyStill
from dxtbx.model import ParallaxCorrectedPxMmStrategy
from dxtbx.model.beam import Beam, BeamFactory
from dxtbx.model.detector import Detector, DetectorFactory


class FormatCBFCspad(FormatCBFMultiTileHierarchyStill):
    """An image reading class CSPAD CBF files"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CSPD CBF format image, i.e. we can
        make sense of it."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_widefile(image_file.encode(), pycbf.MSG_DIGEST)

        cbf_handle.find_category(b"diffrn_detector")
        if cbf_handle.count_rows() > 1:
            return False  # support 1 detector per file for now

        cbf_handle.find_column(b"type")

        return cbf_handle.get_value() == b"CS PAD"

    def _detector(self):
        d = FormatCBFMultiTileHierarchyStill._detector(self)

        try:
            # a header only CBF file will not have a beam object
            beam = self._beam()
        except Exception as e:
            if "CBF_NOTFOUND" in str(e):
                return d
            raise e

        # take into consideration here the thickness of the sensor also the
        # wavelength of the radiation (which we have in the same file...)
        wavelength = beam.get_wavelength()
        thickness = 0.5  # mm, see Hart et al. 2012

        table = attenuation_coefficient.get_table("Si")
        # mu_at_angstrom returns cm^-1
        mu = table.mu_at_angstrom(wavelength) / 10.0  # mu: mm^-1
        t0 = thickness

        for panel in d:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
        return d

    def sync_detector_to_cbf(self, detector=None):
        """If the detector object has been changed, due to refinement or manual repositioning
        in a gui, call this function to synchronize these changes to the underlying CBF handle
        """

        def recursive_sync(cbf, group, cbf_detectors=None, root=False):
            """Walks the hierarchy and synchronizes the dxtbx matrices to the CBF axes."""

            d_mat = group.get_local_d_matrix()
            fast = col((d_mat[0], d_mat[3], d_mat[6])).normalize()
            slow = col((d_mat[1], d_mat[4], d_mat[7])).normalize()
            orig = col((d_mat[2], d_mat[5], d_mat[8]))

            if group.is_panel():
                if cbf.has_sections():
                    # use the pre-mapping
                    cbf_detector = cbf_detectors[group.get_name().encode()]
                else:
                    # figure out which panel number this panel is by finding it in diffrn_data_frame
                    cbf.find_category(b"diffrn_data_frame")
                    cbf.find_column(b"array_id")
                    cbf.find_row(group.get_name())
                    cbf.find_column(b"binary_id")
                    array_num = int(cbf.get_value()) - 1

                    cbf_detector = cbf.construct_detector(array_num)

                axis0 = cbf_detector.get_detector_surface_axes(0)
                axis1 = cbf_detector.get_detector_surface_axes(1)
                assert cbf.get_axis_depends_on(axis0) == axis1
                name = cbf.get_axis_depends_on(axis1)

                # account for the sign change when moving from dxtbx to ImageCIF
                slow = col((slow[0], -slow[1], slow[2]))

                # account for the offset from the center of the sensor to the center
                # of the ASIC.
                orig -= col(cbf.get_axis_offset(axis1))
            else:
                name = group.get_name().encode()

            v3 = fast.cross(slow).normalize()

            if root:
                # use the X, Y, Z settings table to record root detector position instead of axis offsets
                dx = orig[0]
                dy = orig[1]
                distance = orig[2]
                orig = col((0, 0, 0))

            cbf.find_category(b"axis")
            cbf.find_column(b"id")
            while True:
                if cbf.get_value() == name:
                    axis_name = cbf.get_value()
                    break
                cbf.next_row()

            cbf.find_column(b"offset[1]")
            cbf.set_value(str(orig[0]).encode())
            cbf.find_column(b"offset[2]")
            cbf.set_value(str(orig[1]).encode())
            cbf.find_column(b"offset[3]")
            cbf.set_value(str(orig[2]).encode())

            r3 = sqr(
                (
                    fast[0],
                    slow[0],
                    v3[0],
                    fast[1],
                    slow[1],
                    v3[1],
                    fast[2],
                    slow[2],
                    v3[2],
                )
            )

            (
                angle,
                axis,
            ) = r3.r3_rotation_matrix_as_unit_quaternion().unit_quaternion_as_axis_and_angle(
                deg=True
            )
            if angle == 0:
                axis = col((0, 0, 1))

            assert axis.length() > 0

            cbf.find_column(b"vector[1]")
            cbf.set_value(str(axis[0]).encode())
            cbf.find_column(b"vector[2]")
            cbf.set_value(str(axis[1]).encode())
            cbf.find_column(b"vector[3]")
            cbf.set_value(str(axis[2]).encode())

            cbf.set_axis_setting(axis_name, angle, 0)

            if root:
                # synchronize the new root origin
                for axis_id, setting_value in zip(
                    [b"_X", b"_Y", b"_Z"], [dx, dy, distance]
                ):
                    while axis_id not in axis_name:
                        axis_name = cbf.get_axis_depends_on(axis_name)
                    assert cbf.get_axis_type(axis_name) == b"translation"
                    cbf.set_axis_setting(axis_name, setting_value, 0)

            if group.is_group():
                for subgroup in group:
                    recursive_sync(cbf, subgroup, cbf_detectors)

        if detector is None:
            detector = self.get_detector()

        root = detector.hierarchy()
        assert len(root) == 4

        cbf = self._get_cbf_handle()
        if cbf.has_sections():
            """
            when using sections, the order of the panels in the detector object doesn't match the
            order returned by cbflib's construct_detector.  The mapping between them is found
            in the table array_structure list, which maps an array_section_id to axis_set_id,
            which can then be matchedup with the axes of the object returned by construct_detector
            """
            all_cbfdetectors = [cbf.construct_detector(i) for i in range(len(detector))]
            all_panelnames = [panel.get_name().encode() for panel in detector]

            # map the array_section_ids, which match the panel names, to their root axis names
            panel_name_mapping = {}
            cbf.find_category(b"array_structure_list")
            for i in range(cbf.count_rows()):
                cbf.find_column(b"array_section_id")
                name = cbf.get_value()
                if name in all_panelnames and name not in panel_name_mapping.values():
                    cbf.find_column(b"axis_set_id")
                    panel_name_mapping[cbf.get_value()] = name
                cbf.next_row()

            # map the panel names (same as array_section_ids) to specific cbf detector objects
            mapped_detectors = {}
            for cbf_d in all_cbfdetectors:
                root_axis = cbf_d.get_detector_surface_axes(0)
                mapped_detectors[panel_name_mapping[root_axis]] = cbf_d
        else:
            mapped_detectors = None

        recursive_sync(cbf, root, mapped_detectors, True)


class FormatCBFFullStillInMemory(FormatCBFFullStill):
    """Overrides the Format object's init method to accept a cbf handle instead
    of a file name. Used with XFELs when it is desirable to never write
    a file to disk, but to process it only in memory.
    """

    @staticmethod
    def understand(image_file):
        """This class shouldn't be found by the dxtbx Registry. Instead it should
        be instantiated directly as needed
        """
        return False

    def __init__(self, cbf_handle, **kwargs):
        """@param cbf_handle In memory cbf_handle, already initialized"""
        self._goniometer_instance = None
        self._scan_instance = None
        self._detector_instance = None
        self._detector_factory = DetectorFactory
        self._beam_factory = BeamFactory

        self._cbf_handle = cbf_handle
        self._raw_data = None
        try:
            detector_instance = self._detector()
            assert isinstance(detector_instance, Detector)
            self._detector_instance = detector_instance

            beam_instance = self._beam()
            assert isinstance(beam_instance, Beam)
            self._beam_instance = beam_instance

        except Exception:
            # FIXME ideally should not squash the errors here...
            pass
        finally:
            self._end()

    def __getstate__(self):
        # For pickling and copying, don't carry the cbf_handle
        return self._detector_instance, self._beam_instance, self.get_raw_data()

    def __setstate__(self, state):
        # For pickling and copying, don't carry the cbf_handle
        self._goniometer_instance = None
        self._scan_instance = None
        self._detector_instance = state[0]
        self._beam_instance = state[1]
        self._raw_data = state[2]

    def __del__(self):
        # If copied or pickled, may not have _cbf_handle any more
        if hasattr(self, "_cbf_handle") and self._cbf_handle is not None:
            self._cbf_handle.__swig_destroy__(self._cbf_handle)


class FormatCBFMultiTileHierarchyStillInMemory(
    FormatCBFFullStillInMemory, FormatCBFMultiTileHierarchyStill
):
    def get_raw_data(self):
        # need to specify the get_raw_data function needed
        return FormatCBFMultiTileHierarchyStill.get_raw_data(self)


class FormatCBFCspadInMemory(FormatCBFMultiTileHierarchyStillInMemory, FormatCBFCspad):
    """mixin class for in memory cspad images"""



"""An implementation of the SMV image reader for JHSim images."""

from __future__ import annotations

import calendar
import sys
import time

from iotbx.detectors import SMVImage

from dxtbx.format.FormatSMV import FormatSMV


class FormatSMVJHSim(FormatSMV):
    """A class for reading SMV format JHSim images, and correctly constructing
    a model for the experiment from this."""

    # all ADSC detectors generate images with an ADC offset of 40
    # for Mar/Rayonix it is 10
    # Rigaku SMV uses 20, and 5 for image plate formats
    # The images generated for https://bl831.als.lbl.gov/~jamesh/dose_slice/results.html
    # use 1, so if a value is not provided in the header, default to 1
    ADC_OFFSET = 1

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an JHSim SMV format image, i.e. we can
        make sense of it. From JH: "The best way to identify images from any of my
        simulators is to look for BEAMLINE=fake in the header."."""

        size, header = FormatSMV.get_smv_header(image_file)

        if header.get("BEAMLINE", "").lower() == "fake":
            return True
        else:
            return False

    def detectorbase_start(self):
        if not hasattr(self, "detectorbase") or self.detectorbase is None:
            self.detectorbase = SMVImage(self._image_file)
            self.detectorbase.open_file = self.open_file
            self.detectorbase.readHeader()

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header."""

        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        image_pedestal = float(
            self._header_dictionary.get("ADC_OFFSET", self.ADC_OFFSET)
        )
        max_trusted_value = 65535 - image_pedestal
        min_trusted_value = 0

        # interpret beam center conventions
        image_height_mm = pixel_size * image_size[1]
        adxv_beam_center = (beam_x, beam_y)
        cctbx_beam_center = (
            adxv_beam_center[0] + pixel_size,
            image_height_mm - adxv_beam_center[1] + pixel_size,
        )

        # Guess whether this is mimicking a Pilatus, if so set detector type so
        # that spot-finding parameters are appropriate
        if pixel_size == 0.172:
            stype = "SENSOR_PAD"
        else:
            stype = "CCD"

        return self._detector_factory.simple(
            stype,
            distance,
            cctbx_beam_center,
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            (min_trusted_value, max_trusted_value),
            [],
            pedestal=image_pedestal,
        )

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = float(self._header_dictionary["WAVELENGTH"])

        return self._beam_factory.simple(wavelength)

    def _scan(self):
        """Return the scan information for this image."""
        exposure_time = 1
        epoch = None

        # PST, PDT timezones not recognised by default...

        epoch = 0
        try:
            date_str = self._header_dictionary["DATE"]
            date_str = date_str.replace("PST", "").replace("PDT", "")
        except KeyError:
            date_str = ""
        for format_string in ["%a %b %d %H:%M:%S %Y", "%a %b %d %H:%M:%S %Z %Y"]:
            try:
                epoch = calendar.timegm(time.strptime(date_str, format_string))
                break
            except ValueError:
                pass

        # assert(epoch)
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        assert len(self.get_detector()) == 1
        panel = self.get_detector()[0]
        image_size = panel.get_image_size()
        return self._get_endianic_raw_data(size=image_size)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVJHSim.understand(arg))



from __future__ import annotations

import warnings

from dxtbx.model.experiment_list import (
    BeamComparison,
    DetectorComparison,
    FormatChecker,
    GoniometerComparison,
)

__all__ = [
    "BeamComparison",
    "DetectorComparison",
    "FormatChecker",
    "GoniometerComparison",
]

# Remove this module after DIALS 3.17
warnings.warn(
    "dxtbx.datablock is deprecated; please import from dxtbx.experiment_list instead",
    UserWarning,
    stacklevel=2,
)



from __future__ import annotations

import sys

from matplotlib import pylab

from scitbx.array_family import flex

import dxtbx.util
from dxtbx.model import ParallaxCorrectedPxMmStrategy
from dxtbx.model.experiment_list import ExperimentListFactory


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    experiments = ExperimentListFactory.from_args(
        args or sys.argv[1:], check_format=False
    )
    assert len(experiments) == 1
    detectors = experiments.detectors()
    assert len(detectors) == 1
    detector = detectors[0]
    assert len(detector) == 1
    px_mm = detector[0].get_px_mm_strategy()
    assert isinstance(px_mm, ParallaxCorrectedPxMmStrategy)
    print("Mu: %f mm^-1 " % px_mm.mu())
    print("t0: %f mm" % px_mm.t0())

    image_size = detector[0].get_image_size()[::-1]
    xcorr = flex.double(flex.grid(image_size))
    ycorr = flex.double(flex.grid(image_size))
    pixel_size = detector[0].get_pixel_size()
    for j in range(xcorr.all()[0]):
        for i in range(xcorr.all()[1]):
            x1, y1 = detector[0].pixel_to_millimeter((i, j))
            x0, y0 = i * pixel_size[0], j * pixel_size[1]
            xcorr[j, i] = x1 - x0
            ycorr[j, i] = y1 - y0
    vmin = min([flex.min(xcorr), flex.min(ycorr)])
    vmax = max([flex.max(xcorr), flex.max(ycorr)])
    fig, ax = pylab.subplots()
    pylab.subplot(121)
    pylab.imshow(xcorr.as_numpy_array(), interpolation="none", vmin=vmin, vmax=vmax)
    pylab.subplot(122)
    im = pylab.imshow(
        ycorr.as_numpy_array(), interpolation="none", vmin=vmin, vmax=vmax
    )
    fig.subplots_adjust(right=0.8)
    cax = fig.add_axes([0.9, 0.1, 0.03, 0.8])
    fig.colorbar(im, cax=cax)
    pylab.show()


if __name__ == "__main__":
    run()



from __future__ import annotations

import sys

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format, abstract
from dxtbx.format.FormatMultiImage import FormatMultiImage


@abstract
class FormatHDF5(FormatMultiImage, Format):
    def __init__(self, image_file, **kwargs):
        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        FormatMultiImage.__init__(self, **kwargs)
        Format.__init__(self, image_file, **kwargs)

    @staticmethod
    def understand(image_file):
        try:
            with FormatHDF5.open_file(image_file, "rb") as fh:
                return fh.read(8) == b"\211HDF\r\n\032\n"
        except OSError:
            return False


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatHDF5.understand(arg))



from __future__ import annotations

import functools
import os

from scitbx.array_family import flex

from dxtbx.format.Format import Format, abstract
from dxtbx.format.image import ImageBool
from dxtbx.imageset import ImageSequence, ImageSet, ImageSetData, ImageSetLazy
from dxtbx.model import MultiAxisGoniometer


def _add_static_mask_to_iset(format_instance: Format, iset: ImageSet) -> None:
    """Combine any static mask from a Format with the ImageSet's mask"""
    if format_instance is not None:
        static_mask = format_instance.get_static_mask()
        if static_mask is not None:
            if not iset.external_lookup.mask.data.empty():
                for m1, m2 in zip(static_mask, iset.external_lookup.mask.data):
                    m1 &= m2.data()
                iset.external_lookup.mask.data = ImageBool(static_mask)
            else:
                iset.external_lookup.mask.data = ImageBool(static_mask)


class Reader:
    def __init__(self, format_class, filenames, num_images=None, **kwargs):
        self.kwargs = kwargs
        self.format_class = format_class
        assert len(filenames) == 1
        self._filename = filenames[0]
        if num_images is None:
            format_instance = self.format_class.get_instance(
                self._filename, **self.kwargs
            )
            self._num_images = format_instance.get_num_images()
        else:
            self._num_images = num_images

    def nullify_format_instance(self):
        self.format_class._current_instance_ = None
        self.format_class._current_filename_ = None

    def read(self, index):
        format_instance = self.format_class.get_instance(self._filename, **self.kwargs)
        return format_instance.get_raw_data(index)

    def paths(self):
        return [self._filename]

    def __len__(self):
        return self._num_images

    def copy(self, filenames, num_images=None):
        return Reader(self.format_class, filenames, num_images, **self.kwargs)

    def identifiers(self):
        return ["%s-%d" % (self._filename, index) for index in range(len(self))]

    def is_single_file_reader(self):
        return True

    def master_path(self):
        return self._filename


@abstract
class FormatMultiImage(Format):
    def __init__(self, **kwargs):
        pass

    def get_num_images(self):
        raise NotImplementedError

    def get_goniometer(self, index=None):
        return self._goniometer_instance

    def get_detector(self, index=None):
        return self._detector_instance

    def get_beam(self, index=None):
        return self._beam_instance

    def get_spectrum(self, index=None):
        raise NotImplementedError

    def get_scan(self, index=None):
        return self._scan_instance

    def get_raw_data(self, index=None):
        raise NotImplementedError

    def get_detectorbase(self, index=None):
        raise NotImplementedError

    @classmethod
    def get_reader(cls):
        """
        Return a reader class
        """
        return functools.partial(Reader, cls)

    def get_masker(self, goniometer=None):
        if (
            isinstance(goniometer, MultiAxisGoniometer)
            and hasattr(self, "_dynamic_shadowing")
            and self._dynamic_shadowing
        ):
            masker = self.get_goniometer_shadow_masker(goniometer=goniometer)
        else:
            masker = None
        return masker

    @classmethod
    def get_imageset(
        cls,
        filenames,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
        as_sequence=False,
        as_imageset=False,
        single_file_indices=None,
        format_kwargs=None,
        template=None,
        check_format=True,
        lazy=False,
    ):
        """
        Factory method to create an imageset
        """

        if isinstance(filenames, str):
            filenames = [filenames]
        elif len(filenames) > 1:
            assert len(set(filenames)) == 1
            filenames = filenames[0:1]

        # Make filenames absolute
        filenames = [os.path.abspath(x) for x in filenames]

        # Make it a dictionary
        if format_kwargs is None:
            format_kwargs = {}

        # If get_num_images hasn't been implemented, we need indices for number of images
        if cls.get_num_images == FormatMultiImage.get_num_images:
            assert single_file_indices is not None
            assert len(single_file_indices) > 0
            num_images = len(single_file_indices)
        else:
            num_images = None

        # Get the format instance
        assert len(filenames) == 1
        if check_format is True:
            cls._current_filename_ = None
            cls._current_instance_ = None
            format_instance = cls.get_instance(filenames[0], **format_kwargs)
            if hasattr(format_instance, "lazy"):
                lazy = format_instance.lazy
            if num_images is None and not lazy:
                # As we now have the actual format class we can get the number
                # of images from here. This saves having to create another
                # format class instance in the Reader() constructor
                # NOTE: Having this information breaks internal assumptions in
                #       *Lazy classes, so they have to figure this out in
                #       their own time.
                num_images = format_instance.get_num_images()
        else:
            format_instance = None
            if not as_sequence:
                lazy = True

        # Get some information from the format class
        reader = cls.get_reader()(filenames, num_images=num_images, **format_kwargs)

        # Read the vendor type
        if check_format is True:
            vendor = format_instance.get_vendortype()
        else:
            vendor = ""

        # Get the format kwargs
        params = format_kwargs

        # Check if we have a sequence

        # Make sure only 1 or none is set
        assert [as_imageset, as_sequence].count(True) < 2
        if as_imageset:
            is_sequence = False
        elif as_sequence:
            is_sequence = True
        else:
            if scan is None and format_instance is None:
                raise RuntimeError(
                    """
          One of the following needs to be set
            - as_imageset=True
            - as_sequence=True
            - scan
            - check_format=True
      """
                )
            if scan is None:
                test_scan = format_instance.get_scan()
            else:
                test_scan = scan
            if test_scan is not None:
                is_sequence = True
            else:
                is_sequence = False

        assert not (as_sequence and lazy), "No lazy support for sequences"

        if single_file_indices is not None:
            assert len(single_file_indices)
            single_file_indices = flex.size_t(single_file_indices)

        # Create an imageset or sequence
        if not is_sequence:
            # Use imagesetlazy
            # Setup ImageSetLazy and just return it. No models are set.
            if lazy:
                try:
                    iset = ImageSetLazy(
                        ImageSetData(
                            reader=reader,
                            masker=None,
                            vendor=vendor,
                            params=params,
                            format=cls,
                        ),
                        indices=single_file_indices,
                    )
                    _add_static_mask_to_iset(format_instance, iset)
                    return iset
                except RuntimeError:
                    # psana2
                    if not (
                        format_instance.get_num_images() == 0
                        or len(single_file_indices) == 0
                    ):
                        raise
                    return None

            # Create the imageset
            iset = ImageSet(
                ImageSetData(
                    reader=reader, masker=None, vendor=vendor, params=params, format=cls
                ),
                indices=single_file_indices,
            )

            if single_file_indices is None:
                single_file_indices = range(format_instance.get_num_images())

            # If any are None then read from format
            if not all((beam, detector, goniometer, scan)):
                # Get list of models
                num_images = format_instance.get_num_images()
                beam = [None] * num_images
                detector = [None] * num_images
                goniometer = [None] * num_images
                scan = [None] * num_images
                for i in single_file_indices:
                    beam[i] = format_instance.get_beam(i)
                    detector[i] = format_instance.get_detector(i)
                    goniometer[i] = format_instance.get_goniometer(i)
                    scan[i] = format_instance.get_scan(i)

            # Set the list of models
            for i, index in enumerate(single_file_indices):
                iset.set_beam(beam[index], i)
                iset.set_detector(detector[index], i)
                iset.set_goniometer(goniometer[index], i)
                iset.set_scan(scan[index], i)

        else:
            # Get the template
            template = filenames[0]

            # Check indices are sequential
            if single_file_indices is not None:
                assert all(
                    i + 1 == j
                    for i, j in zip(single_file_indices[:-1], single_file_indices[1:])
                )
                num_images = len(single_file_indices)
            else:
                num_images = format_instance.get_num_images()

            # Check the scan makes sense - we must want to use <= total images
            if scan is not None:
                assert scan.get_num_images() <= num_images

            # If any are None then read from format
            if beam is None:
                beam = format_instance.get_beam()
            if detector is None:
                detector = format_instance.get_detector()
            if goniometer is None:
                goniometer = format_instance.get_goniometer()
            if scan is None:
                scan = format_instance.get_scan()

            # Create the masker
            if format_instance is not None:
                masker = format_instance.get_masker(goniometer=goniometer)
            else:
                masker = None

            isetdata = ImageSetData(
                reader=reader,
                masker=masker,
                vendor=vendor,
                params=params,
                format=cls,
                template=template,
            )

            # Create the sequence
            iset = ImageSequence(
                isetdata,
                beam=beam,
                detector=detector,
                goniometer=goniometer,
                scan=scan,
                indices=single_file_indices,
            )

        _add_static_mask_to_iset(format_instance, iset)

        return iset



from __future__ import annotations

import xml
from typing import Dict, Generic, List, Tuple, TypeVar

import numpy as np
from scipy.constants import Planck, m_n
from scipy.interpolate import interp1d

import cctbx.array_family.flex as flex

Shape = TypeVar("Shape")
DType = TypeVar("DType")


class Array(np.ndarray, Generic[Shape, DType]):
    pass


vec3float = Array["3", float]
vec2float = Array["2", float]
vec3int = Array["3", int]
vec2int = Array["2", int]


def wavelength_from_tof(
    distance: float | flex.double, tof: float | flex.double
) -> float | flex.double:
    """
    distance (m)
    tof (s)

    return (A)
    """
    return ((Planck * tof) / (m_n * distance)) * 10**10


def tof_from_wavelength(
    distance: float | flex.double, wavelength: float | flex.double
) -> float | flex.double:
    """
    wavelength (A)
    return (s)
    """

    wavelength = wavelength * 10**-10

    return (wavelength * m_n * distance) / Planck


def frame_to_tof_interpolator(frames: list[float], tof: list[float]) -> interp1d:
    """
    ToF can vary nonlinearly with frame number.
    A cubic spline is used to better account for these cases.
    """
    assert min(frames) >= 0
    assert min(tof) >= 0
    assert len(frames) == len(tof)
    assert all(i < j for i, j in zip(frames, frames[1:]))
    assert all(i < j for i, j in zip(tof, tof[1:]))
    return interp1d(frames, tof, kind="cubic")


def tof_to_frame_interpolator(tof: list[float], frames: list[float]) -> interp1d:
    """
    ToF can vary nonlinearly with frame number.
    A cubic spline is used to better account for these cases.
    """
    assert min(frames) >= 0
    assert min(tof) >= 0
    assert len(frames) == len(tof)
    assert all(i < j for i, j in zip(frames, frames[1:]))
    assert all(i < j for i, j in zip(tof, tof[1:]))
    return interp1d(tof, frames, kind="cubic")


class InstrumentDefinitionReader:
    """
    Class to obtain instrument information in dxtbx format from an
    Instrument Definition File as described here:
    https://docs.mantidproject.org/nightly/concepts/InstrumentDefinitionFile.html#instrumentdefinitionfile

    Example usage:
    >>> import h5py
    >>> h5_file = h5py.File("path/to/nexus_file")
    >>> xml_string = h5_file["path/to/xml_file"]
    >>> import xml.etree.ElementTree as ET
    >>> xml_file = ET.fromstring(xml_string)
    >>> from dxtx.model.tof_helpers import InstrumentDefinitionReader
    >>> reader = InstrumentDefinitionReader()
    >>> origins, fast_axes, slow_axes = reader.get_dials_detector_geometry(xml_file)
    """

    class Panel:
        """
        Class to store panel properties in spherical coordinates
        """

        def __init__(
            self,
            idx: int,
            centre_origin_in_m: float,
            gam_in_deg: float,
            nu_in_deg: float,
            num_pixels: vec2int,
            pixel_size_in_m: vec2float,
            x_orientation: vec2int = np.array((1, 0)),
            y_orientation: vec2int = np.array((0, 1)),
        ) -> None:
            self.idx = idx
            self.centre_origin_in_m = centre_origin_in_m
            self.gam_in_deg = gam_in_deg
            self.nu_in_deg = nu_in_deg
            self.num_pixels = num_pixels
            self.pixel_size_in_m = pixel_size_in_m
            self.x_orientation = x_orientation
            self.y_orientation = y_orientation

        def panel_size_in_m(self) -> vec2float:
            panel_size = []
            for i in range(len(self.num_pixels)):
                panel_size.append(self.num_pixels[i] * self.pixel_size_in_m[i])
            return np.array(panel_size)

        def orientations_flipped(self) -> bool:
            abs_x = tuple([abs(i) for i in self.x_orientation])
            abs_y = tuple([abs(i) for i in self.y_orientation])
            return abs_x == (0, 1) and abs_y == (1, 0)

        def orientation_direction_flipped(self) -> bool:
            return sum(self.x_orientation) == sum(self.y_orientation) == -1

        def __repr__(self) -> None:
            return f"idx: {self.idx} \n \
            centre origin (m): {self.centre_origin_in_m} \n \
            gam (deg): {self.gam_in_deg} \n \
            nu (deg): {self.nu_in_deg} \n \
            num pixels: {self.num_pixels} \n \
            pixel size (m): {self.pixel_size_in_m} \n \
            x_orientation: {self.x_orientation} \n \
            y_orientation: {self.y_orientation}"

    def is_panel(self, tree_component: xml.etree.ElementTree.Element) -> bool:
        if "type" not in tree_component.attrib:
            return False
        return (
            "panel" in tree_component.attrib["type"]
            and "location" in tree_component[0].tag
        )

    def get_rotation_vals(
        self, rot: xml.etree.ElementTree.Element
    ) -> Tuple[float, vec3float]:
        if "rot" in rot.attrib:
            val = float(rot.attrib["rot"])
        elif "val" in rot.attrib:
            val = float(rot.attrib["val"])
        else:
            return None
        try:
            x = int(rot.attrib["axis-x"])
            y = int(rot.attrib["axis-y"])
            z = int(rot.attrib["axis-z"])
        except KeyError:  # if no axes given axis-z is assumed
            x = 0
            y = 0
            z = 1
        return (val, (x, y, z))

    def get_rotations(
        self,
        line: xml.etree.ElementTree.Element,
        rotations: List[Tuple[float, vec3float]],
    ) -> List[Tuple[float, vec3float]]:
        rotation_vals = self.get_rotation_vals(line)
        if rotation_vals is not None:
            rotations.append(self.get_rotation_vals(line))
        try:
            return self.get_rotations(line[0], rotations=rotations)
        except IndexError:
            return rotations

    def x_project_v(self, x: vec3float, v: vec3float) -> Dict[str, vec3float]:
        """Project x onto v, returning parallel and perpendicular components
        >> d = xProject(x, v)
        >> np.allclose(d['par'] + d['perp'], x)
        True
        """
        par = self.x_par_v(x, v)
        perp = x - par
        return {"par": par, "perp": perp}

    def rotate_about(self, a: vec3float, b: vec3float, theta: float) -> vec3float:
        if np.isclose(abs(np.dot(a, b)), 1):
            return a
        """Rotate vector a about vector b by theta radians."""

        proj = self.x_project_v(a, b)
        w = np.cross(b, proj["perp"])
        unit_w = w / np.linalg.norm(w)
        return (
            proj["par"]
            + proj["perp"] * np.cos(theta)
            + np.linalg.norm(proj["perp"]) * unit_w * np.sin(theta)
        )

    def x_par_v(self, x: vec3float, v: vec3float) -> vec3float:
        """Project x onto v. Result will be parallel to v."""
        return np.dot(x, v) / np.dot(v, v) * v

    def x_perp_v(self, x: vec3float, v: vec3float) -> vec3float:
        """Component of x orthogonal to v. Result is perpendicular to v."""
        return x - self.x_par_v(x, v)

    def get_panel_axes(
        self,
        xml_file: xml.etree.ElementTree.Element,
        init_fast_axis: vec3float,
        init_slow_axis: vec3float,
    ) -> Tuple[vec3float, vec3float, vec3float]:
        def get_rot_axis(
            raw_rot_axis: vec3float,
            fast_axis: vec3float,
            slow_axis: vec3float,
            z_axis: vec3float,
            init_fast_axis: vec3float,
            init_slow_axis: vec3float,
            init_z_axis: vec3float,
        ) -> vec3float:
            if abs(np.dot(raw_rot_axis, init_fast_axis)) == 1:
                return fast_axis
            if abs(np.dot(raw_rot_axis, init_slow_axis)) == 1:
                return slow_axis
            if abs(np.dot(raw_rot_axis, init_z_axis)) == 1:
                return z_axis
            raise NotImplementedError

        def get_axes(
            rotations: List[Tuple[float, vec3float]],
            init_fast_axis: vec3float,
            init_slow_axis: vec3float,
            init_z_axis: vec3float,
        ) -> Tuple[Tuple[vec3float], Tuple[vec3float]]:
            fast_axis = init_fast_axis
            slow_axis = init_slow_axis
            z_axis = init_z_axis
            for rot in rotations:
                raw_rot_axis = np.array(rot[1])
                rot_val = np.deg2rad(rot[0])
                rot_axis = get_rot_axis(
                    raw_rot_axis,
                    fast_axis,
                    slow_axis,
                    z_axis,
                    init_fast_axis,
                    init_slow_axis,
                    init_z_axis,
                )

                fast_axis = self.rotate_about(fast_axis, rot_axis, rot_val)
                slow_axis = self.rotate_about(slow_axis, rot_axis, rot_val)
                z_axis = self.rotate_about(z_axis, rot_axis, rot_val)
            return fast_axis, slow_axis, z_axis

        slow_axes = []
        fast_axes = []
        init_z_axis = np.cross(init_fast_axis, init_slow_axis)
        for child in xml_file:
            if self.is_panel(child):
                panel = child[0]
                rotations = []
                rotations = self.get_rotations(
                    line=panel,
                    rotations=rotations,
                )
                axes = get_axes(rotations, init_fast_axis, init_slow_axis, init_z_axis)
                fast_axes.append(tuple(axes[0]))
                slow_axes.append(tuple(axes[1]))
        return fast_axes, slow_axes

    def shift_origin_centre_to_top_left(
        self,
        centre_origin_in_m: vec3float,
        fast_axis: vec3float,
        slow_axis: vec3float,
        panel_size_in_m: vec2float,
    ) -> vec3float:
        # Assumes fast and slow axes are unit vectors

        slow_axis = slow_axis * panel_size_in_m[1] * 0.5
        fast_axis = fast_axis * panel_size_in_m[0] * 0.5
        return centre_origin_in_m - slow_axis - fast_axis

    def rotations_to_spherical_coordinates(
        self,
        zeroth_pixel_origin: vec2float,
        rotations: Tuple[Tuple[float, vec3int], ...],
    ) -> vec2float:
        """
        Corrects rotations for zeroth_pixel_origin and returns
        the gam and nu angles in spherical coordinates.
        """

        gam = rotations[0][0]

        try:
            nu = rotations[1][0]
        except IndexError:
            nu = 0

        xstart, ystart = zeroth_pixel_origin
        if np.sign(xstart) == np.sign(ystart):
            if abs(nu) > 0:
                nu *= -1
            else:
                gam *= -1

        elif abs(nu) > 0:
            gam -= 180

        return gam, nu

    def get_panel_types(
        self, xml_file: xml.etree.ElementTree.Element
    ) -> Dict[str, float | int]:
        def is_panel_settings(tree_element: xml.etree.ElementTree.Element) -> bool:
            required_fields = [
                "xstart",
                "ystart",
                "xpixels",
                "ypixels",
                "xstep",
                "ystep",
            ]
            for i in required_fields:
                if i not in tree_element.attrib:
                    return False

            return True

        panel_types = {}

        for child in xml_file:
            if is_panel_settings(child):
                key = child.attrib["name"]
                xstart = float(child.attrib["xstart"])
                ystart = float(child.attrib["ystart"])
                xpixels = int(child.attrib["xpixels"])
                ypixels = int(child.attrib["ypixels"])
                xpixel_size = abs(float(child.attrib["xstep"]))
                ypixel_size = abs(float(child.attrib["ystep"]))
                panel_types[key] = {
                    "xstart": xstart,
                    "ystart": ystart,
                    "xpixels": xpixels,
                    "ypixels": ypixels,
                    "xpixel_size": xpixel_size,
                    "ypixel_size": ypixel_size,
                }

        return panel_types

    def get_panels(self, xml_file: xml.etree.ElementTree.Element) -> Tuple[Panel]:
        def panel_name_to_idx(name: str) -> int:
            return int("".join(filter(str.isdigit, name)))

        def get_cartesian_origin(panel_attrib):
            try:
                x = float(panel_attrib["x"])
                y = float(panel_attrib["y"])
                z = float(panel_attrib["z"])
            except KeyError:  # assume spherical coordinates
                r = float(panel_attrib["r"])
                t = np.deg2rad(float(panel_attrib["t"]))
                p = np.deg2rad(float(panel_attrib["p"]))
                x = r * np.sin(t) * np.cos(p)
                y = r * np.sin(t) * np.sin(p)
                z = r * np.cos(t)
            return x, y, z

        def get_rotations_from_origin(
            panel_attrib: Dict[str, str],
        ) -> Tuple[float, float]:
            return float(panel_attrib["t"]), float(panel_attrib["p"])

        def get_rotation_vals(
            rot: xml.etree.ElementTree.Element,
        ) -> Tuple[float, vec3float]:
            val = float(rot.attrib["val"])
            try:
                x = int(rot.attrib["axis-x"])
                y = int(rot.attrib["axis-y"])
                z = int(rot.attrib["axis-z"])
            except KeyError:  # if no axes given axis-z is assumed
                x = 0
                y = 0
                z = 1
            return (val, (x, y, z))

        def get_rotations(
            line: xml.etree.ElementTree.Element,
            rotations: List[Tuple[float, vec3float]],
        ) -> List[Tuple[float, vec3float]]:
            rotations.append(get_rotation_vals(line))
            try:
                return get_rotations(line[0], rotations=rotations)
            except IndexError:
                return rotations

        def get_panel_orentation(name: str) -> Tuple[vec2int, vec2int]:
            """
            This is a hack for SXD, where the Mantid SXD_Definition.xml
            does not seem to identify SXD panel 1 as upsidedown.
            """

            if name == "bank1":
                return np.array((0, -1)), np.array((-1, 0))
            else:
                return np.array((1, 0)), np.array((0, 1))

        panels = []
        panel_types = self.get_panel_types(xml_file)

        for child in xml_file:
            if self.is_panel(child):
                panel = child[0]
                name = panel.attrib["name"]
                idx = panel_name_to_idx(name)
                x, y, z = get_cartesian_origin(panel.attrib)
                rotation_start = panel[0]
                rotations = []
                rotations = get_rotations(
                    line=rotation_start,
                    rotations=rotations,
                )

                panel_type = child.attrib["type"]
                panel_info = panel_types[panel_type]

                zeroth_pixel_origin = (panel_info["xstart"], panel_info["ystart"])
                gam_in_deg, nu_in_deg = self.rotations_to_spherical_coordinates(
                    zeroth_pixel_origin=zeroth_pixel_origin, rotations=rotations
                )
                try:
                    gam_in_deg, nu_in_deg = get_rotations_from_origin(panel.attrib)
                except KeyError:
                    gam_in_deg = 0
                    nu_in_deg = 0

                num_pixels = (
                    panel_info["xpixels"],
                    panel_info["ypixels"],
                )

                pixel_size_in_m = (panel_info["xpixel_size"], panel_info["ypixel_size"])

                x_or, y_or = get_panel_orentation(name=name)

                panels.append(
                    InstrumentDefinitionReader.Panel(
                        idx=idx,
                        centre_origin_in_m=(x, y, z),
                        gam_in_deg=gam_in_deg,
                        nu_in_deg=nu_in_deg,
                        num_pixels=num_pixels,
                        pixel_size_in_m=pixel_size_in_m,
                        x_orientation=x_or,
                        y_orientation=y_or,
                    )
                )

        return tuple(panels)

    def panel_idx_to_name(self, idx: int) -> str:
        return "bank" + str(idx)

    def get_panel_names(self, xml_file: xml.etree.ElementTree.Element) -> Tuple[str]:
        panels = self.get_panels(xml_file)
        panel_names = [self.panel_idx_to_name(i.idx) for i in panels]
        return panel_names

    def get_num_panels(self, xml_file: xml.etree.ElementTree.Element) -> int:
        return len(self.get_panels(xml_file))

    def get_dials_detector_geometry(
        self, xml_file: xml.etree.ElementTree.Element
    ) -> Tuple[Tuple[vec3float], Tuple[vec3float], Tuple[vec3float]]:
        # Get panel data in spherical coordinates
        panels = self.get_panels(xml_file)
        panel_dict = {self.panel_idx_to_name(i.idx): i for i in panels}

        # Get panel axes
        init_slow_axis = np.array((0, 1, 0))
        init_fast_axis = np.array((1, 0, 0))

        fast_axes, slow_axes = self.get_panel_axes(
            xml_file, init_fast_axis, init_slow_axis
        )

        # Get panel origins
        origins = []
        count = 0
        for name in panel_dict:
            new_panel = panel_dict[name]
            fast_axis = np.array(fast_axes[count])
            slow_axis = np.array(slow_axes[count])
            top_left_origin_in_m = self.shift_origin_centre_to_top_left(
                centre_origin_in_m=new_panel.centre_origin_in_m,
                fast_axis=fast_axis,
                slow_axis=slow_axis,
                panel_size_in_m=new_panel.panel_size_in_m(),
            )
            top_left_origin_in_mm = np.array([i * 1000 for i in top_left_origin_in_m])

            origins.append(tuple(top_left_origin_in_mm))
            count += 1

        origins = tuple(tuple(float(v) for v in vec) for vec in origins)
        fast_axes = tuple(tuple(float(v) for v in vec) for vec in fast_axes)
        slow_axes = tuple(tuple(float(v) for v in vec) for vec in slow_axes)
        return origins, fast_axes, slow_axes



from __future__ import annotations

import json
import math
import os
import sys
from urllib.parse import urlparse

import libtbx


class AutoEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, libtbx.AutoType):
            return "Auto"
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def encode_output_as_utf8() -> None:
    """
    Ensures stdout and stderr are UTF-8 encoded.
    This avoids crashes on Windows when printing unicode where the output is
    redirected away from the console into a pipe or file
    """
    if hasattr(encode_output_as_utf8, "done"):
        return

    if sys.stdout.encoding.lower() != "utf-8":
        sys.stdout.reconfigure(encoding="utf-8")  # type: ignore
        sys.stderr.reconfigure(encoding="utf-8")  # type: ignore
    encode_output_as_utf8.done = True  # type: ignore


def format_float_with_standard_uncertainty(value, standard_uncertainty, minimum=1e-12):
    """
    Formats a float, including the uncertainty in its value.

    Parameters
    ----------
    value : float
    standard_uncertainty : float
    minimum : float

    Returns
    -------
    str

    Examples
    --------
    >>> format_float_with_standard_uncertainty(5e-3, 1e-3)
    '0.0050(10)'
    >>> format_float_with_standard_uncertainty(5e-3, 1e-6)
    '0.0050000(10)'
    """
    if standard_uncertainty <= minimum:
        dp = -int(math.log10(minimum))
        return str(round(value, dp))
    precision = -int(round(math.log10(standard_uncertainty)))
    if precision > -1:
        su = standard_uncertainty * math.pow(10, precision)
        if round(su, 1) < 2:
            su *= 10
            precision += 1
        return "{value:.{precision}f}({irsu})".format(
            value=value, precision=precision, irsu=int(round(su))
        )
    else:
        precision += 1
        su = int(round(standard_uncertainty, precision))
        fmt_str = "%.0f(%i)"
        return fmt_str % (round(value, precision), su)


def show_mask_info(expt_list):
    """Print the number of masked pixels for each module in the detector,
    for each experiment in the input list."""
    for i in expt_list.imagesets():
        d = i.get_detector()
        m = i.get_mask(0)
        print("---- ----")
        print(d)
        for j, _m in enumerate(m):
            print(f"Module {j} has {_m.count(False)} masked pixels of {_m.size()}")


def get_url_scheme(url) -> str:
    """Extract the URL scheme from the string url, respecting Windows file paths"""

    return urlparse(str(url)).scheme if "://" in str(url) else ""


def ersatz_uuid4() -> str:
    """Generate an ersatz UUID4 i.e. random 128 bit UUID."""

    h = "%032x" % int.from_bytes(os.urandom(16), byteorder="little")
    return "%s-%s-%s-%s-%s" % (h[:8], h[8:12], h[12:16], h[16:20], h[20:])



from __future__ import annotations

import os

import pycbf

import libtbx.phil
from scitbx.array_family import flex

from dxtbx.model.scan_helpers import scan_helper_image_files

try:
    from ..dxtbx_model_ext import Scan
except ModuleNotFoundError:
    from dxtbx_model_ext import Scan  # type: ignore

scan_phil_scope = libtbx.phil.parse(
    """
  scan
    .expert_level = 1
    .short_caption = "Scan overrides"
  {

    image_range = None
      .type = ints(size=2)
      .help = "Override the image range"
      .short_caption = "Image range"

    extrapolate_scan = False
      .type = bool
      .help = "When overriding the image range, extrapolate exposure and epoch information from existing images"
      .short_caption = "Extrapolate scan"

    oscillation = None
      .type = floats(size=2)
      .help = "Override the image oscillation"
      .short_caption = "Oscillation"

    batch_offset = None
      .type = int(value_min=0)
      .help = "Override the batch offset"
      .short_caption = "Batch offset"
  }
"""
)


class ScanFactory:
    """A factory for scan instances, to help with constructing the classes
    in a set of common circumstances."""

    @staticmethod
    def from_phil(params, reference=None):
        """
        Generate a scan model from phil parameters

        """
        if reference is None:
            if params.scan.image_range is None and params.scan.oscillation is None:
                return None
            if params.scan.image_range is None:
                raise RuntimeError("No image range set")
            if params.scan.oscillation is None:
                raise RuntimeError("No oscillation set")
            scan = Scan(params.scan.image_range, params.scan.oscillation)
        else:
            scan = reference

            if params.scan.image_range is not None:
                most_recent_image_index = (
                    scan.get_image_range()[1] - scan.get_image_range()[0]
                )
                if scan.has_property("oscillation"):
                    scan.set_oscillation(
                        scan.get_image_oscillation(params.scan.image_range[0])
                    )
                scan.set_image_range(params.scan.image_range)
                if (
                    params.scan.extrapolate_scan
                    and (params.scan.image_range[1] - params.scan.image_range[0])
                    > most_recent_image_index
                ):
                    exposure_times = scan.get_exposure_times()
                    epochs = scan.get_epochs()
                    exposure_time = exposure_times[most_recent_image_index]
                    epoch_correction = epochs[most_recent_image_index]
                    for i in range(
                        most_recent_image_index + 1,
                        params.scan.image_range[1] - params.scan.image_range[0] + 1,
                    ):
                        exposure_times[i] = exposure_time
                        epoch_correction += exposure_time
                        epochs[i] = epoch_correction
                    scan.set_epochs(epochs)
                    scan.set_exposure_times(exposure_times)

            if params.scan.oscillation is not None:
                scan.set_oscillation(params.scan.oscillation)

        if params.scan.batch_offset is not None:
            scan.set_batch_offset(params.scan.batch_offset)

        return scan

    @staticmethod
    def from_dict(d, t=None):
        """Convert the dictionary to a scan model

        Params:
            d The dictionary of parameters
            t The template dictionary to use

        Returns:
            The scan model
        """

        def add_properties_table(scan_dict, num_images):
            """
            Handles legacy case before Scan had a properties table.
            Moves oscillation, epochs, and exposure times to a properties
            table and adds this to scan_dict.
            """

            properties = {}
            if scan_dict:
                if "oscillation" in scan_dict:
                    if num_images == 1:
                        properties["oscillation_width"] = [scan_dict["oscillation"][1]]
                        properties["oscillation"] = [scan_dict["oscillation"][0]]

                    else:
                        osc = scan_dict["oscillation"]
                        properties["oscillation"] = [
                            osc[0] + (osc[1] - osc[0]) * i for i in range(num_images)
                        ]
                    del scan_dict["oscillation"]
                if "exposure_time" in scan_dict:
                    properties["exposure_time"] = scan_dict["exposure_time"]
                    del scan_dict["exposure_time"]
                if "epochs" in scan_dict:
                    properties["epochs"] = scan_dict["epochs"]
                    del scan_dict["epochs"]
            scan_dict["properties"] = make_properties_table_consistent(
                properties, num_images
            )
            return scan_dict

        def make_properties_table_consistent(properties, num_images):
            """
            Handles legacy case before Scan had a properties table.
            Ensures oscillation, epochs, and exposure times have the same length.
            """

            if not properties:
                return properties

            property_length = len(next(iter(properties.values())))
            all_same_length = all(
                len(lst) == property_length for lst in properties.values()
            )
            if all_same_length and property_length == num_images:
                return properties

            if "oscillation" in properties:
                assert len(properties["oscillation"]) > 0

                if num_images == 1 and "oscillation_width" not in properties:
                    assert len(properties["oscillation"]) > 1
                    properties["oscillation_width"] = [properties["oscillation"][1]]
                    properties["oscillation"] = [properties["oscillation"][0]]
                elif num_images > 1:
                    osc_0 = properties["oscillation"][0]
                    if "oscillation_width" in properties:
                        osc_1 = properties["oscillation_width"][0]
                        del properties["oscillation_width"]
                    else:
                        assert len(properties["oscillation"]) > 1
                        osc_1 = (
                            properties["oscillation"][1] - properties["oscillation"][0]
                        )
                    properties["oscillation"] = [
                        osc_0 + osc_1 * i for i in range(num_images)
                    ]

            if "exposure_time" in properties:
                assert len(properties["exposure_time"]) > 0

                # Assume same exposure time for each image
                properties["exposure_time"] = [
                    properties["exposure_time"][0] for i in range(num_images)
                ]

            if "epochs" in properties:
                assert len(properties["epochs"]) > 0
                # If 1 epoch, assume increasing by epochs[0]
                # Else assume increasing as epochs[1] - epochs[0]
                if len(properties["epochs"]) == 1:
                    properties["epochs"] = [
                        properties["epochs"][0] + properties["epochs"][0] * i
                        for i in range(num_images)
                    ]
                else:
                    diff = properties["epochs"][1] - properties["epochs"][0]
                    properties["epochs"] = [
                        properties["epochs"][0] + i * diff for i in range(num_images)
                    ]
            return properties

        if d is None and t is None:
            return None
        joint = t.copy() if t else {}

        # Accounting for legacy cases where t or d does not
        # contain properties dict
        num_images = None
        if "image_range" in d:
            num_images = 1 + d["image_range"][1] - d["image_range"][0]
        elif "image_range" in joint:
            num_images = 1 + joint["image_range"][1] - joint["image_range"][0]

        if "properties" in joint and "properties" in d:
            properties = t["properties"].copy()
            properties.update(d["properties"])
            joint.update(d)
            joint["properties"] = properties
        elif "properties" in d:
            joint = add_properties_table(joint, num_images)
            d_copy = d.copy()
            joint["properties"].update(d_copy["properties"])
            joint["properties"] = make_properties_table_consistent(
                joint["properties"], num_images
            )
            del d_copy["properties"]
            joint.update(d_copy)
        elif "properties" in joint:
            d = add_properties_table(d, num_images)
            d_copy = d.copy()
            joint["properties"].update(d_copy["properties"])
            joint["properties"] = make_properties_table_consistent(
                joint["properties"], num_images
            )
            del d_copy["properties"]
            joint.update(d_copy)
        else:
            joint.update(d)

        if "properties" not in d and not isinstance(joint["exposure_time"], list):
            joint["exposure_time"] = [joint["exposure_time"]]
        joint.setdefault("batch_offset", 0)  # backwards compatibility 20180205
        joint.setdefault("valid_image_ranges", {})  # backwards compatibility 20181113

        return Scan.from_dict(joint)

    @staticmethod
    def make_scan(
        image_range, exposure_times, oscillation, epochs, batch_offset=0, deg=True
    ):
        if not isinstance(exposure_times, list):
            num_images = image_range[1] - image_range[0] + 1
            exposure_times = [exposure_times for i in range(num_images)]
        else:
            num_images = image_range[1] - image_range[0] + 1
            num_exp = len(exposure_times)
            if num_exp != num_images:
                if num_exp == 0:
                    exposure_times = [0 for i in range(num_images)]
                else:
                    exposure_times = exposure_times.extend(
                        [exposure_times[-1] for i in range(num_images - num_exp)]
                    )

        epoch_list = [epochs[j] for j in sorted(epochs)]

        return Scan(
            tuple(map(int, image_range)),
            tuple(map(float, oscillation)),
            flex.double(list(map(float, exposure_times))),
            flex.double(list(map(float, epoch_list))),
            batch_offset,
            deg,
        )

    @staticmethod
    def make_scan_from_properties(image_range, properties, batch_offset=0, deg=True):
        return Scan(tuple(map(int, image_range)), properties, batch_offset, deg)

    @staticmethod
    def single_file(filename, exposure_times, osc_start, osc_width, epoch):
        """Construct an scan instance for a single image."""
        index = scan_helper_image_files.image_to_index(os.path.split(filename)[-1])
        if epoch is None:
            epoch = 0.0

        # if the oscillation width is negative at this stage it is almost
        # certainly an artefact of the omega end being 0 when the omega start
        # angle was ~ 360 so it would be ~ -360 - see dxtbx#378
        if osc_width < -180:
            osc_width += 360

        return ScanFactory.make_scan(
            (index, index), exposure_times, (osc_start, osc_width), {index: epoch}
        )

    @staticmethod
    def imgCIF(cif_file):
        """Initialize a scan model from an imgCIF file."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_file(cif_file, pycbf.MSG_DIGEST)

        return ScanFactory.imgCIF_H(cif_file, cbf_handle)

    @staticmethod
    def imgCIF_H(cif_file, cbf_handle):
        """Initialize a scan model from an imgCIF file handle, where it is
        assumed that the file has already been read."""

        exposure = cbf_handle.get_integration_time()
        timestamp = cbf_handle.get_timestamp()[0]

        gonio = cbf_handle.construct_goniometer()
        try:
            angles = tuple(gonio.get_rotation_range())
        except Exception as e:
            if str(e).strip() == "CBFlib Error(s): CBF_NOTFOUND":
                # probably a still shot -> no scan object
                return None
            raise

        # xia2-56 handle gracefully reverse turning goniometers - this assumes the
        # rotation axis is correctly inverted in the goniometer factory
        if angles[1] < 0:
            angles = -angles[0], -angles[1]

        index = scan_helper_image_files.image_to_index(cif_file)

        gonio.__swig_destroy__(gonio)

        return ScanFactory.make_scan(
            (index, index), exposure, angles, {index: timestamp}
        )

    @staticmethod
    def add(scans):
        """Sum a list of scans wrapping the slightly clumsy idiomatic method:
        sum(scans[1:], scans[0])."""
        return sum(scans[1:], scans[0])

    @staticmethod
    def search(filename):
        """Get a list of files which appear to match the template and
        directory implied by the input filename. This could well be used
        to get a list of image headers to read and hence construct scans
        from."""

        template, directory = scan_helper_image_files.image_to_template_directory(
            filename
        )

        indices = scan_helper_image_files.template_directory_to_indices(
            template, directory
        )

        return [
            scan_helper_image_files.template_directory_index_to_image(
                template, directory, index
            )
            for index in indices
        ]



"""
Root class for still shots.  A still shot has no goniomter and no
scan in their model, as these constructs are not meaningful.
"""

from __future__ import annotations

from dxtbx.format.Format import Format, abstract
from dxtbx.model.beam import Beam
from dxtbx.model.detector import Detector


@abstract
class FormatStill(Format):
    def setup(self):
        """Read the image file, construct the information which we will be
        wanting about the experiment from this. N.B. in your implementation
        of this you will probably want to make use of the static methods
        below and probably add some format parsing code too. Please also keep
        in mind that your implementation may be further subclassed by
        someone else.

        Do not create scan or or goniometer objects"""
        self._start()
        try:
            detector_instance = self._detector()
            assert isinstance(detector_instance, Detector)
            self._detector_instance = detector_instance

            beam_instance = self._beam()
            assert isinstance(beam_instance, Beam)
            self._beam_instance = beam_instance

        except Exception:
            # FIXME ideally should not squash the errors here...
            pass
        finally:
            self._end()

    def _goniometer(self):
        """Not sensible for still shot data"""

        return None

    def _scan(self):
        """Not sensible for still shot data"""

        return None



"""
A registry class to handle Format classes and provide lists of them when
this is useful for i.e. identifying the best tool to read a given range
of image formats.
"""

from __future__ import annotations

import importlib.metadata
import os
import sys
import typing
from collections.abc import Callable

from dxtbx.util import get_url_scheme

if typing.TYPE_CHECKING:
    from dxtbx.format.Format import Format


def get_format_class_for(format_class_name: str) -> type[Format]:
    """Return the named format class
    :param format_class_name: Name of the format class
    :return: The (uninstantiated) class object
    """
    return get_format_class_index()[format_class_name][0]()


def get_format_class_index() -> dict[str, tuple[Callable[[], type[Format]], list[str]]]:
    """Return a dictionary of all known format classes.
    :return: A dictionary containing entries
             {format_class_name: (format_class_factory_function, [base_class_names])}
             The factory function takes no arguments and returns an
             uninstantiated format class. This avoids importing all
             format classes.
    """
    if not hasattr(get_format_class_index, "cache"):
        class_index = {}
        if sys.version_info < (3, 10):
            # REMOVE when python 3.10 is minimum.
            # This was was deprecated in 3.10, and removed in 3.12.
            format_ep = importlib.metadata.entry_points()["dxtbx.format"]
        else:
            format_ep = importlib.metadata.entry_points(group="dxtbx.format")
        for e in format_ep:
            if ":" in e.name:
                format_name, base_classes_str = e.name.split(":", 1)
                base_classes = tuple(base_classes_str.split(","))
            else:
                format_name, base_classes = e.name, ()
            class_index[format_name] = (e.load, base_classes)
        setattr(get_format_class_index, "cache", class_index)
    register = get_format_class_index.cache.copy()  # type: ignore
    return register


def get_format_class_dag() -> dict[str, list[str]]:
    """Return a directed acyclical graph of the format classes.
    :return: A dictionary with entries
             {format class name: [subformat class names]}
    """
    if not hasattr(get_format_class_dag, "cache"):
        index = {
            name: class_info[1] for name, class_info in get_format_class_index().items()
        }
        dag: dict[str, list[str]] = {}
        for name in index:
            for parent in index[name]:
                dag.setdefault(parent, []).append(name)
        for key in dag:
            dag[key].sort()
        setattr(get_format_class_dag, "cache", dag)
    dag = get_format_class_dag.cache.copy()  # type: ignore
    return dag


_format_dag: dict[str, list[str]] = get_format_class_dag()


def get_format_class_for_file(
    image_file: str | bytes | os.PathLike, format_hint: str | None = None
) -> type[Format] | None:
    """Find the best format handler in the registry for given image file
    :param image_file: A string containing the file path to an image
    :param format_hint: An optional string of a format class name that should
                        be checked first
    :return: An uninstantiated format class, if a matching one was found,
             otherwise None
    """

    # Grab the scheme from this URI
    scheme = get_url_scheme(image_file)

    image_file_str = os.fspath(image_file)

    # If a format hint was given then find all paths through the DAG leading
    # to this format class. Create a set containing all format class names
    # on those paths (priority_formats), and use the information whether a
    # format name is in the set to prioritise it during the format search.
    priority_formats = set()
    if format_hint:
        priority_candidates = {format_hint}
        dagset = {f: set(subf) for f, subf in _format_dag.items()}
        while priority_candidates:
            priority_formats.update(priority_candidates)
            priority_candidates = {
                f
                for f, subf in dagset.items()
                if priority_candidates.intersection(subf)
            }

    def format_sort(format_name):
        return (0 if format_name in priority_formats else 1, format_name)

    def recurse(format_name, image_file):
        # Recursively check whether any of the children understand image_file,
        # in which case they are preferred over the parent format.
        for child in sorted(_format_dag.get(format_name, []), key=format_sort):
            format_class = get_format_class_for(child)
            if scheme in format_class.schemes and format_class.understand(image_file):
                return recurse(child, image_file)
        return get_format_class_for(format_name)

    # Starting at "Format" and using any potential prioritisation information
    # look for any path through the DAG of formats, stopping at the first
    # accepting leaf node
    for format in sorted(_format_dag["Format"], key=format_sort):
        format_class = get_format_class_for(format)
        if scheme in format_class.schemes and format_class.understand(image_file_str):
            return recurse(format, image_file_str)

    return None



from __future__ import annotations

import os

import h5py
import numpy as np
import pytest

from scitbx.array_family import flex

from dxtbx.format.FormatNXmxDLS import FormatNXmxDLS
from dxtbx.format.FormatNXmxDLS16M import FormatNXmxDLS16M
from dxtbx.masking import SmarGonShadowMasker
from dxtbx.model.experiment_list import ExperimentListFactory

dials = pytest.importorskip("dials")
pytest.importorskip("h5py")


@pytest.mark.parametrize(
    "master_h5",
    [
        "/dls/i04/data/2021/cm28182-4/TestProteinaseK/protK3/protK3_1_master.h5",
        "/dls/i04/data/2021/cm28182-4/TestProteinaseK/protK3/protK3_1.nxs",
    ],
)
def test_rotation_scan_i04(master_h5):
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames(
        [master_h5], format_kwargs={"dynamic_shadowing": True}
    )
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatNXmxDLS16M

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == (0.075, 0.075)
    assert panel.get_image_size() == (4148, 4362)
    assert panel.get_trusted_range() == (0, 22726)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_origin() == pytest.approx(
        (-158.21930645001163, 166.30914199665585, -201.83999382350086)
    )
    assert panel.get_distance() == 201.83999382350086

    assert len(gonio.get_axes()) == 3
    expected_axes = ((1, 0, 0), (0, 0, -1), (1, 0, 0))
    for a1, a2 in zip(gonio.get_axes(), expected_axes):
        assert a1 == pytest.approx(a2, abs=5e-2)
    assert gonio.get_scan_axis() == 2

    assert pytest.approx(scan.get_oscillation()) == (0, 0.1)
    assert scan.get_image_range() == (1, 3600)

    assert beam.get_wavelength() == pytest.approx(0.979499130984)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))


@pytest.mark.parametrize(
    "master_h5",
    [
        "/dls/i04/data/2022/cm31106-5/20221118/SmarGon/rotation_calibration4/protK15_O45_C45_P45_1_master.h5",
        "/dls/i04/data/2022/cm31106-5/20221118/SmarGon/rotation_calibration4/protK15_O45_C45_P45_1.nxs",
    ],
)
def test_rotation_scan_i04_2022_run_5(master_h5):
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames(
        [master_h5], format_kwargs={"dynamic_shadowing": True}
    )
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatNXmxDLS16M

    gonio = imageset.get_goniometer()
    assert list(gonio.get_angles()) == pytest.approx([45.0, 45.0, 45.0])
    assert list(gonio.get_axes().as_double()) == pytest.approx(
        [1.0, -0.0037, 0.002, -0.0046, 0.0372, -0.9993, 1.0, 0.0, 0.0]
    )
    assert list(gonio.get_names()) == ["phi", "chi", "omega"]
    assert imageset.has_dynamic_mask()
    masker = imageset.masker()
    assert isinstance(masker, SmarGonShadowMasker)
    assert masker.get_mask(imageset.get_detector(), 0)[0].count(False) == 0
    masker.get_mask(imageset.get_detector(), 50)[0].count(False) == 486717
    assert masker.get_mask(imageset.get_detector(), 100)[0].count(False) == 1110799


@pytest.mark.skip(reason="Data erased from CM visit")
@pytest.mark.parametrize(
    "master_h5",
    [
        "/dls/i04/data/2020/cm26459-3/20200617/bs/lres_1_master.h5",
        "/dls/i04/data/2020/cm26459-3/20200617/bs/lres_1.nxs",
    ],
)
def test_masked_i04_32bit(master_h5):
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert flex.max(imageset[0][0]) != 0x7FFFFFFF


@pytest.mark.skip(reason="Data erased from CM visit")
@pytest.mark.parametrize(
    "master_h5",
    [
        "/dls/i03/data/2020/cm26458-3/20200617/test_1_master.h5",
        "/dls/i03/data/2020/cm26458-3/20200617/test_1.nxs",
    ],
)
def test_masked_i03_16bit(master_h5):
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert flex.min(imageset[0][0]) == -1.0
    assert flex.max(imageset[0][0]) != 0xFFFF


def test_grid_scan_i04():
    master_h5 = "/dls/i04/data/2019/cm23004-1/20190109/Eiger/grid/Thaum/Thau_5/Thau_5_1_master.h5"
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatNXmxDLS16M

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == (0.075, 0.075)
    assert panel.get_image_size() == (4148, 4362)
    assert panel.get_trusted_range() == (0, 65535)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_origin() == pytest.approx(
        (-167.44717577120824, 172.46833023184868, -350.0)
    )
    assert panel.get_distance() == 350

    assert len(gonio.get_axes()) == 3
    expected_axes = ((1, 0, 0), (0, 0, -1), (1, 0, 0))
    for a1, a2 in zip(gonio.get_axes(), expected_axes):
        assert a1 == pytest.approx(a2, abs=5e-2)
    # assert gonio.get_scan_axis() == 2

    if scan:
        osc = scan.get_oscillation()
        assert osc[0] == osc[1]

    assert beam.get_wavelength() == pytest.approx(0.979499)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))


@pytest.mark.xfail(
    raises=AssertionError, reason="https://github.com/cctbx/dxtbx/issues/13"
)
def test_screening(dials_data):
    master_h5 = dials_data("thaumatin_eiger_screen") / "Therm_6_1_master.h5"
    assert FormatNXmxDLS16M.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    assert len(expts) == 3
    imagesets = expts[0].imageset
    assert imagesets[0].get_format_class() == FormatNXmxDLS16M


@pytest.mark.parametrize("beamline", ["I03", "I04"])
def test_understand(beamline, tmp_path):
    # See https://jira.diamond.ac.uk/browse/MXGDA-3624
    nxs = tmp_path / "data.nxs"
    with h5py.File(nxs, mode="w") as fh:
        entry = fh.create_group("entry")
        instrument = entry.create_group("instrument")
        instrument.attrs["short_name"] = np.bytes_(f"DLS {beamline}")
        name = instrument.create_dataset(
            "name", data=np.bytes_(f"DIAMOND BEAMLINE {beamline}")
        )
        name.attrs["short_name"] = np.bytes_(f"DLS {beamline}")
    assert FormatNXmxDLS16M.understand(nxs)
    assert FormatNXmxDLS.understand(nxs)


@pytest.mark.parametrize("beamline", ["I03", "I04"])
def test_understand_legacy(beamline, tmp_path):
    # See https://jira.diamond.ac.uk/browse/MXGDA-3624
    nxs = tmp_path / "data.nxs"
    with h5py.File(nxs, mode="w") as fh:
        entry = fh.create_group("entry")
        instrument = entry.create_group("instrument")
        instrument.attrs["short_name"] = np.bytes_(f"{beamline}")
        name = instrument.create_dataset("name", data=np.bytes_(f"{beamline}"))
        name.attrs["short_name"] = np.bytes_(f"{beamline}")
    assert FormatNXmxDLS16M.understand(nxs)
    assert FormatNXmxDLS.understand(nxs)


def test_do_not_understand_name_none(tmp_path):
    nxs = tmp_path / "data.nxs"
    with h5py.File(nxs, mode="w") as fh:
        entry = fh.create_group("entry")
        entry.create_group("instrument")
    assert not FormatNXmxDLS16M.understand(nxs)


def test_do_not_understand_i24(tmp_path):
    nxs = tmp_path / "data.nxs"
    with h5py.File(nxs, mode="w") as fh:
        entry = fh.create_group("entry")
        instrument = entry.create_group("instrument")
        instrument.attrs["short_name"] = np.bytes_("DLS I24")
        name = instrument.create_dataset("name", data=np.bytes_("DIAMOND BEAMLINE I24"))
        name.attrs["short_name"] = np.bytes_("DLS I24")
    assert not FormatNXmxDLS16M.understand(nxs)



from __future__ import annotations

from typing import Any

from scitbx.array_family import flex

from .dxtbx_model_ext import BeamBase, Detector, MultiAxisGoniometer, Panel

def is_inside_polygon(polygon: flex.vec2_double, x: float, y: float) -> bool: ...
def mask_untrusted_rectangle(
    mask: flex.bool, x0: int, x1: int, y0: int, y1: int
) -> None: ...
def mask_untrusted_circle(
    mask: flex.bool, xc: float, yc: float, radius: float
) -> None: ...
def mask_untrusted_polygon(mask: flex.bool, polygon: flex.vec2_double) -> None: ...
def mask_untrusted_resolution_range(
    mask: flex.bool, beam: BeamBase, panel: Panel, d_min: float, d_max: float
) -> None: ...

class GoniometerShadowMasker:
    def __init__(
        self,
        goniometer: MultiAxisGoniometer,
        extrama_at_datum: flex.vec3_double,
        axis: flex.size_t,
        invert_mask: bool = False,
    ) -> None: ...
    def extrema_at_scan_angle(self, scan_angle: float) -> flex.vec3_double: ...
    def set_goniometer_angles(dxtbx, scitbx) -> Any: ...
    def project_extrema(
        self, detector: Detector, scan_angle: float
    ) -> flex.vec2_double: ...
    def get_mask(self, detector: Detector, scan_angle: float) -> tuple[flex.bool]: ...

class SmarGonShadowMasker(GoniometerShadowMasker):
    def __init__(self, goniometer: MultiAxisGoniometer) -> None: ...
    def extrema_at_scan_angle(self, scan_angle: float) -> flex.vec3_double: ...



"""
Convert a NXmx-format NeXus file to a set of CBF-format image files.

Note that this tool does not produce full imgCIF-format files, only
DECTRIS-style mini-CBF files consisting of a plain text simplified
header and the binary compressed image data.  The simplified header
does not contain a full description of the experiment geometry and some
metadata, including the detector orientation, are unspecified.  As
such, you may need to use these mini-CBF files in conjunction with a
known instrument model in a separate file.
"""

from __future__ import annotations

import argparse
import sys
from pathlib import Path

import dxtbx.util.dlsnxs2cbf

parser = argparse.ArgumentParser(description=__doc__, prog="dxtbx.dlsnxs2cbf")
parser.add_argument(
    "nexus_file", metavar="nexus-file", help="Input NeXus file.", type=Path
)
parser.add_argument(
    "-o",
    "--output-directory",
    help="Directory in which to store the CBF image files.  Defaults to the current "
    "working directory.",
    type=Path,
    default=Path(),
)
parser.add_argument(
    "-p",
    "--parents",
    help="Create all parents of the output directory if they do not already exist.  "
    "By default, an exception will be raised if part of the output path is missing.",
    action="store_true",
)
parser.add_argument(
    "-t",
    "--name-template",
    help="Template for the CBF file name stem.  The output filenames will be "
    "constructed by appending an image number and '.cbf' file extension to this "
    "template.  For example, the template 'image_' will result in files named like "
    "'image_001.cbf'.  To name the files by image number alone, e.g. '0001.cbf', "
    "pass an empty template string with '-t \"\"'.  The default template is the stem "
    "of the input NeXus file name, with an added trailing underscore.",
)
parser.add_argument(
    "-n",
    "--number-of-digits",
    help="The number of digits to use for the image number label in the CBF file "
    "names.  If this exceeds the minimal necessary number of digits to accommodate "
    "the number of the last CBF image, the image number label will be padded with "
    "zeros.  If an insufficient number is specified, the minimal necessary number of "
    "digits will be used instead, which is the default behaviour.",
    type=int,
    default=0,
)


def run(args: list[str] | None = None):
    dxtbx.util.encode_output_as_utf8()

    args = parser.parse_args(args)

    try:
        args.output_directory.mkdir(parents=args.parents, exist_ok=True)
    except FileNotFoundError as e:
        sys.exit(e)

    dxtbx.util.dlsnxs2cbf.make_cbf(
        args.nexus_file,
        args.output_directory,
        args.name_template,
        args.number_of_digits,
    )


if __name__ == "__main__":
    run()



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for example on APS ID19 SN 458 and 914
which have reversed phi.
"""

from __future__ import annotations

import sys

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSNAPSID19(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument numbers 458 and 914
    from the APS ID19 beamline."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 458 or 914."""

        # check this is detector serial number 458 or 914

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        if int(header["DETECTOR_SN"]) not in (458, 914):
            return False

        return True

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )

        key = [
            s for s in self._header_dictionary if s.endswith("_SPATIAL_BEAM_POSITION")
        ][0]

        beam_x, beam_y = (
            float(f) * pixel_size for f in self._header_dictionary[key].split()
        )

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, (image_size[1] * pixel_size) - beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            self._adsc_trusted_range(),
            [],
            gain=self._adsc_module_gain(),
        )

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header."""

        return self._goniometer_factory.single_axis_reverse()

    def _scan(self):
        """Return the scan information for this image. There may be
        no timestamps in there..."""

        exposure_time = float(self._header_dictionary["TIME"])
        epoch = 0
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCSN.understand(arg))



from __future__ import annotations

import numpy as np
import pytest
from scipy.ndimage import rotate
from scipy.stats import norm

from dxtbx.util import ersatz_uuid4
from dxtbx.util import format_float_with_standard_uncertainty as ffwsu
from dxtbx.util.rotate_and_average import rotate_and_average


def test_format_float_with_standard_uncertainty():
    """Test format_float_with_standard_uncertainty."""

    assert ffwsu(21.234567, 0.0013) == "21.2346(13)"
    assert ffwsu(21.234567, 0.0023) == "21.235(2)"
    assert ffwsu(12345, 45) == "12340(40)"
    assert ffwsu(12.3, 1.2) == "12.3(12)"
    assert ffwsu(-0.2451, 0.8135) == "-0.2(8)"
    assert ffwsu(1.234, 0.196) == "1.2(2)"
    assert ffwsu(1.234, 0.193) == "1.23(19)"
    assert ffwsu(90, 0) == "90"
    assert ffwsu(90.0, 0) == "90.0"
    assert ffwsu(90 + 1e-14, 0) == "90.0"


def test_ersatz_uuid4():
    """Test ersatz UUID4 behaviour:
    - (pseudo)randomness and uniqueness
    - shape e.g. 00000000-0000-0000-0000-000000000000
    - content e.g. hex digits"""

    uuids = [ersatz_uuid4() for j in range(1000)]

    assert len(uuids) == len(set(uuids))

    for uuid in uuids:
        bits = uuid.split("-")
        assert tuple(len(b) for b in bits) == (8, 4, 4, 4, 12)
        for b in bits:
            assert set(b) <= set("0123456789abcdef")


def test_rotate_and_average():
    # create a pseudo-spectrum by multiplying two gaussians, to make
    # final image of 20,200, then rotate by 10 degrees
    x = np.linspace(-50, 49, 200)
    fx = norm.pdf(x, 10, 30)  # mean 10, stddev 30
    y = np.linspace(-10, 9, 20)
    fy = norm.pdf(y, 2, 5)  # mean 2, stddev 5

    xpart = np.vstack([fx] * 20)
    ypart = np.vstack([fy] * 200).transpose()

    img = xpart * ypart  # 2d gaussian
    img = rotate(img, angle=10, reshape=True)

    # unrotate and average along the main axis
    x, y = rotate_and_average(img, -10, deg=True)

    # apply ROI
    roi = slice(15, -15)
    x = x[roi]
    y = y[roi]

    unrotated_mean = np.average(
        range(img.shape[1])[roi], weights=np.mean(img, axis=0)[roi]
    )
    rotated_mean = np.average(x, weights=y)

    assert 110.305114 == pytest.approx(unrotated_mean)
    assert 104.311567 == pytest.approx(rotated_mean)



"""
Experimental format for TIA .ser files used by some FEI microscopes. See
https://personal.ntu.edu.sg/cbb/info/TIAformat/index.html
"""

from __future__ import annotations

import os
import re
import struct
import xml.etree.ElementTree as ET
from math import sqrt
from pathlib import Path

import numpy as np

from boost_adaptbx.boost.python import streambuf
from scitbx.array_family import flex

import dxtbx.ext
from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format
from dxtbx.format.FormatMultiImage import FormatMultiImage
from dxtbx.model.beam import Probe


# The read_emi and _parseEntry_emi functions are taken from openNCEM project
# and modified here under the terms of the MIT license.
#
# Copyright 2019 the Ncempy developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def read_emi(filename):
    """Read the meta data from an emi file.
    Parameters
    ----------
        filename: str or pathlib.Path
            Path to the emi file.
    Returns
    -------
        : dict
            Dictionary of experimental metadata stored in the EMI file.
    """

    # check filename type
    if isinstance(filename, str):
        pass
    elif isinstance(filename, Path):
        filename = str(filename)
    else:
        raise TypeError("Filename is supposed to be a string or pathlib.Path")

    # Format.open provides transparent decompression
    with Format.open_file(filename, "rb") as f_emi:
        emi_data = f_emi.read()

    # dict to store _emi stuff
    _emi = {}

    # need anything readable from <ObjectInfo> to </ObjectInfo>
    # collect = False
    # data = b''
    # for line in f_emi:
    #    if b'<ObjectInfo>' in line:
    #        collect = True
    #    if collect:
    #        data += line.strip()
    #    if b'</ObjectInfo>' in line:
    #        collect = False

    # close the file
    # f_emi.close()

    metaStart = emi_data.find(b"<ObjectInfo>")
    metaEnd = emi_data.find(
        b"</ObjectInfo>"
    )  # need to add len('</ObjectInfo>') = 13 to encompass this final tag

    root = ET.fromstring(emi_data[metaStart : metaEnd + 13])

    # strip of binary stuff still around
    # data = data.decode('ascii', errors='ignore')
    # matchObj = re.search('<ObjectInfo>(.+?)</ObjectInfo', data)
    # try:
    #    data = matchObj.group(1)
    # except:
    #    raise RuntimeError('Could not find _emi metadata in specified file.')

    # parse metadata as xml
    # root = ET.fromstring('<_emi>' + data + '</_emi>')

    # single items
    _emi["Uuid"] = root.findtext("Uuid")
    _emi["AcquireDate"] = root.findtext("AcquireDate")
    _emi["Manufacturer"] = root.findtext("Manufacturer")
    _emi["DetectorPixelHeight"] = root.findtext("DetectorPixelHeight")
    _emi["DetectorPixelWidth"] = root.findtext("DetectorPixelWidth")

    # Microscope Conditions
    grp = root.find("ExperimentalConditions/MicroscopeConditions")

    for elem in grp:
        _emi[elem.tag] = _parseEntry_emi(elem.text)

    # Experimental Description
    grp = root.find("ExperimentalDescription/Root")

    for elem in grp:
        _emi["{} [{}]".format(elem.findtext("Label"), elem.findtext("Unit"))] = (
            _parseEntry_emi(elem.findtext("Value"))
        )

    # AcquireInfo
    grp = root.find("AcquireInfo")

    for elem in grp:
        _emi[elem.tag] = _parseEntry_emi(elem.text)

    # DetectorRange
    grp = root.find("DetectorRange")

    for elem in grp:
        _emi["DetectorRange_" + elem.tag] = _parseEntry_emi(elem.text)

    return _emi


def _parseEntry_emi(value):
    """Auxiliary function to parse string entry to int, float or np.bytes_().
    Parameters
    ----------
        value : str
            String containing an int, float or string.
    Returns
    -------
        : int or float or str
            Entry value as int, float or string.
    """

    # try to parse as int
    try:
        p = int(value)
    except ValueError:
        # if not int, then try float
        try:
            p = float(value)
        except ValueError:
            # if neither int nor float, stay with string
            p = np.bytes_(str(value))

    return p


class FormatSER(Format):
    @staticmethod
    def understand(image_file):
        try:
            with FormatSER.open_file(image_file, "rb") as fh:
                tag = fh.read(14)
        except OSError:
            return False

        # File should be little endian
        if tag[0:2] != b"II":
            return False

        # SeriesID: 0x0197 indicates ES Vision Series Data File
        if struct.unpack("<H", tag[2:4])[0] != 0x0197:
            return False

        # SeriesVersion: 0x0210 or 0x0220
        if struct.unpack("<H", tag[4:6])[0] not in (0x0210, 0x0220):
            return False

        # DataTypeID: 0x4122 if elements are 2D arrays
        if struct.unpack("<I", tag[6:10])[0] != 0x4122:
            return False

        # TagTypeID: 0x4142 or 0x4152
        if struct.unpack("<I", tag[10:])[0] not in (0x4142, 0x4152):
            return False

        return True

    @staticmethod
    def _read_metadata(image_file):
        hd = {}
        with FormatSER.open_file(image_file, "rb") as f:
            f.seek(4)
            version = struct.unpack("<H", f.read(2))[0]

            # version specific details
            if version == 528:
                size = 4
                int_fmt = "I"
            else:
                size = 8
                int_fmt = "Q"

            f.seek(14)
            hd["TotalNumberElements"] = struct.unpack("<I", f.read(4))[0]
            hd["ValidNumberElements"] = struct.unpack("<I", f.read(4))[0]
            hd["OffsetArrayOffset"] = struct.unpack("<" + int_fmt, f.read(size))[0]
            hd["NumberDimensions"] = struct.unpack("<I", f.read(4))[0]

            dimension_array = []
            for i in range(hd["NumberDimensions"]):
                d = {}
                d["DimensionSize"] = struct.unpack("<I", f.read(4))[0]
                d["CalibrationOffset"] = struct.unpack("<d", f.read(8))[0]
                d["CalibrationDelta"] = struct.unpack("<d", f.read(8))[0]
                d["CalibrationElement"] = struct.unpack("<I", f.read(4))[0]
                d["DescriptionLength"] = struct.unpack("<I", f.read(4))[0]
                d["Description"] = f.read(d["DescriptionLength"])
                d["UnitsLength"] = struct.unpack("<I", f.read(4))[0]
                d["Units"] = f.read(d["UnitsLength"])
                dimension_array.append(d)
            hd["DimensionArray"] = dimension_array

            f.seek(hd["OffsetArrayOffset"])
            hd["DataOffsetArray"] = struct.unpack(
                "<" + int_fmt * hd["TotalNumberElements"],
                f.read(size * hd["TotalNumberElements"]),
            )
            hd["TagOffsetArray"] = struct.unpack(
                "<" + int_fmt * hd["TotalNumberElements"],
                f.read(size * hd["TotalNumberElements"]),
            )

            # get expected image size by reading metadata for the first image
            f.seek(hd["DataOffsetArray"][0] + 42)
            hd["ArraySizeX"] = struct.unpack("<I", f.read(4))[0]
            hd["ArraySizeY"] = struct.unpack("<I", f.read(4))[0]

        try:
            emi = read_emi(image_file.replace(".ser", ".emi"))
            hd.update(emi)
        except FileNotFoundError:
            pass

        return hd

    def _start(self):
        """Open the image file and read useful metadata into an internal dictionary"""
        self._header_dictionary = self._read_metadata(self._image_file)

    def _goniometer(self):
        """Dummy goniometer, 'vertical' as the images are viewed. The handedness
        is set to be appropriate for some datasets obtained from eBIC"""

        return self._goniometer_factory.known_axis((0, -1, 0))

    def _detector(self):
        """Dummy detector"""

        image_size = (
            self._header_dictionary["ArraySizeX"],
            self._header_dictionary["ArraySizeY"],
        )

        # Assumptions relevant for either Ceta or Falcon
        pixel_size = 0.014, 0.014
        binning = 1
        if image_size == (2048, 2048):
            pixel_size = 0.028, 0.028
            binning = 2
        binning = self._header_dictionary.get("Binning", binning)
        saturation = 8000 * binning**2
        trusted_range = (-1000, saturation)

        distance = float(self._header_dictionary.get("Camera length [m]", 2)) * 1000

        # Set beam to the centre of the image
        beam_centre = [(p * i) / 2 for p, i in zip(pixel_size, image_size)]
        d = self._detector_factory.simple(
            "PAD",
            distance,
            beam_centre,
            "+x",
            "-y",
            pixel_size,
            image_size,
            trusted_range,
        )

        # Assume Ceta-D gain unless known to be a Falcon (this neglects standard
        # Ceta gain, which is said to be about 7.0).
        if b"falcon" in self._header_dictionary.get("CameraNamePath", b"").lower():
            gain = 1.0
        else:
            gain = 26.0

        for p in d:
            p.set_gain(gain)
        return d

    def _beam(self):
        """Unpolarized beam, default energy 200 keV unless in metadata"""

        def cal_wavelength(V0):
            h = 6.626e-34  # Js, Planck's constant
            m = 9.109e-31  # kg, electron mass
            e = 1.6021766208e-19  # C, electron charge
            c = 3e8  # m/s^2, speed

            # Default to e- wavelength at 200 keV if voltage set to zero
            if V0 == 0:
                V0 = 200000
            return (
                h / sqrt(2 * m * e * V0 * (1 + e * V0 / (2 * m * c * c))) * 1e10
            )  # return wavelength in Angstrom

        wavelength = cal_wavelength(
            self._header_dictionary.get("AcceleratingVoltage", 200000)
        )
        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )

    def _get_raw_data(self, index):
        data_offset = self._header_dictionary["DataOffsetArray"][index]
        with FormatSER.open_file(self._image_file, "rb") as f:
            f.seek(data_offset)
            d = {}
            d["CalibrationOffsetX"] = struct.unpack("<d", f.read(8))[0]
            d["CalibrationDeltaX"] = struct.unpack("<d", f.read(8))[0]
            d["CalibrationElementX"] = struct.unpack("<I", f.read(4))[0]
            d["CalibrationOffsetY"] = struct.unpack("<d", f.read(8))[0]
            d["CalibrationDeltaY"] = struct.unpack("<d", f.read(8))[0]
            d["CalibrationElementY"] = struct.unpack("<I", f.read(4))[0]
            d["DataType"] = struct.unpack("<H", f.read(2))[0]

            if d["DataType"] == 6:
                read_pixel = dxtbx.ext.read_int32
            elif d["DataType"] == 5:
                read_pixel = dxtbx.ext.read_int16
            elif d["DataType"] == 3:
                read_pixel = dxtbx.ext.read_uint32
            elif d["DataType"] == 2:
                read_pixel = dxtbx.ext.read_uint16
            elif d["DataType"] == 1:
                read_pixel = dxtbx.ext.read_uint8
            else:
                raise RuntimeError(f"Image {index + 1} data is of an unsupported type")

            d["ArraySizeX"] = struct.unpack("<I", f.read(4))[0]
            d["ArraySizeY"] = struct.unpack("<I", f.read(4))[0]
            nelts = d["ArraySizeX"] * d["ArraySizeY"]
            raw_data = read_pixel(streambuf(f), nelts)

        # Check image size is as expected (same as the first image)
        if d["ArraySizeX"] != self._header_dictionary["ArraySizeX"]:
            raise RuntimeError(f"Image {index + 1} has an unexpected array size in X")
        if d["ArraySizeY"] != self._header_dictionary["ArraySizeY"]:
            raise RuntimeError(f"Image {index + 1} has an unexpected array size in Y")

        image_size = (d["ArraySizeX"], d["ArraySizeY"])
        raw_data.reshape(flex.grid(image_size[1], image_size[0]))

        return raw_data


class FormatSERimages(FormatSER):
    @staticmethod
    def understand(image_file):
        with FormatSER.open_file(image_file, "rb") as fh:
            fh.seek(18)
            nimages = struct.unpack("<I", fh.read(4))[0]
        return nimages == 1

    def _scan(self):
        """Dummy scan for this image"""

        fname = os.path.split(self._image_file)[-1]
        # assume that the final number before the extension is the image number
        s = fname.split("_")[-1].split(".")[0]
        try:
            index = int(re.match(".*?([0-9]+)$", s).group(1))
        except AttributeError:
            index = 1
        exposure_times = 0.0
        frame = index - 1
        # Dummy scan with a 0.5 deg image
        oscillation = (frame * 0.5, 0.5)
        epochs = [0]
        return self._scan_factory.make_scan(
            (index, index), exposure_times, oscillation, epochs, deg=True
        )

    def get_raw_data(self):
        return self._get_raw_data(0)


class FormatSERstack(FormatMultiImage, FormatSER):
    @staticmethod
    def understand(image_file):
        with FormatSER.open_file(image_file, "rb") as fh:
            fh.seek(18)
            nimages = struct.unpack("<I", fh.read(4))[0]
        return nimages > 1

    def __init__(self, image_file, **kwargs):
        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        FormatMultiImage.__init__(self, **kwargs)
        Format.__init__(self, image_file, **kwargs)

    def _scan(self):
        """Dummy scan for this stack"""

        nframes = self.get_num_images()
        image_range = (1, nframes)
        exposure_times = 0.0
        oscillation = (0, 0.5)
        epochs = [0] * nframes

        return self._scan_factory.make_scan(
            image_range, exposure_times, oscillation, epochs, deg=True
        )

    def get_num_images(self):
        return self._header_dictionary["ValidNumberElements"]

    def get_goniometer(self, index=None):
        return Format.get_goniometer(self)

    def get_detector(self, index=None):
        return Format.get_detector(self)

    def get_beam(self, index=None):
        return Format.get_beam(self)

    def get_scan(self, index=None):
        if index is None:
            return Format.get_scan(self)
        else:
            scan = Format.get_scan(self)
            return scan[index]

    def get_image_file(self, index=None):
        return Format.get_image_file(self)

    def get_raw_data(self, index):
        return self._get_raw_data(index)



"""
An implementation of the CBF image reader for Eiger images for beamlines
with a horizontal, but reversed rotation axis.
"""

from __future__ import annotations

import sys

from dxtbx.format.FormatCBFMiniEiger import FormatCBFMiniEiger


class FormatCBFMiniEigerReverse(FormatCBFMiniEiger):
    """A class for reading mini CBF format Eiger images for beamlines with
    a horizontal reversed rotation axis."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Eiger mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniEiger.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "# Detector" in record and "Eiger" in record:
                if "S/N E-32-0111" in record:
                    # S/N E-32-0111: SSRF BL17U1
                    return True

        return False

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        return self._goniometer_factory.single_axis_reverse()


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEigerReverse.understand(arg))



"""
Experimental format for the X-Spectrum LAMBDA detector
http://www.x-spectrum.de/
"""

from __future__ import annotations

import sys

import h5py

from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.Format import Format
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.model import ParallaxCorrectedPxMmStrategy


class FormatHDF5Lambda(FormatHDF5):
    @staticmethod
    def understand(image_file):
        try:
            tag = FormatHDF5.open_file(image_file, "rb").read(8)
        except OSError:
            return False

        # check that this is a HDF5 file (should not have got here if not
        # anyway...)

        if tag != b"\211HDF\r\n\032\n":
            return False

        with h5py.File(image_file, "r") as h5_handle:
            try:
                desc = h5_handle["entry/instrument/detector/description"]
            except KeyError:
                return False

            try:
                if b"Lambda" in desc[()][0]:
                    return True
            except TypeError:
                return False

        return False

    def _start(self):
        self._h5_handle = h5py.File(self.get_image_file(), "r")

    def _goniometer(self):
        """Dummy goniometer - assume vertical (EMBl-HH P14)"""

        return self._goniometer_factory.known_axis((0, 1, 0))

    def _detector(self):
        """Partly dummy detector"""
        # Get the detector geometry
        entry = self._h5_handle["entry"]
        instrument = entry["instrument"]
        detector = instrument["detector"]

        # Initialise detector frame - origin at 0,0,0
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, 1.0, 0.0))
        orig = matrix.col((0.0, 0.0, 0.0))

        # Get the pixel and image size
        pixel_size = (
            1.0e-3 * float(detector["x_pixel_size"][0]),
            1.0e-3 * float(detector["y_pixel_size"][0]),
        )
        layout = detector["layout"][0].split(b"X")
        image_size = int(layout[0]), int(layout[1])
        trusted_range = (0, int(detector["saturation_value"][0]))
        thickness = float(detector["sensor_thickness"][0]) / 1000.0
        material = detector["sensor_material"][0].decode()

        # Make the detector
        detector = self._detector_factory.make_detector(
            "PAD",
            fast,
            slow,
            orig,
            pixel_size,
            image_size,
            trusted_range,
            name="Panel",
            thickness=thickness,
            material=material,
        )

        # At the moment, beam is a dummy object because wavelength is not set in
        # the header. Therefore, the px<-->mm strategy will generally be
        # incorrect. Set it anyway, to override later.
        beam = self._beam()
        wavelength = beam.get_wavelength()

        # this will fail for undefined composite materials
        table = attenuation_coefficient.get_table(material)

        # mu_at_angstrom returns cm^-1, but need mu in mm^-1
        mu = table.mu_at_angstrom(wavelength) / 10.0

        for panel in detector:
            panel.set_mu(mu)
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, thickness))

        return detector

    def _beam(self):
        """Dummy beam"""

        wavelength = 1.0
        return self._beam_factory.simple(wavelength)

    def _scan(self):
        """Dummy scan"""

        entry = self._h5_handle["entry"]
        nframes = int(entry["instrument/detector/collection/frame_numbers"][0])
        image_range = (1, nframes)
        exposure_times = 0.0
        oscillation = (0, 1)
        epochs = [0] * nframes

        return self._scan_factory.make_scan(
            image_range, exposure_times, oscillation, epochs, deg=True
        )

    def get_num_images(self):
        detector = self._h5_handle["entry/instrument/detector"]
        data = detector["data"]
        return data.shape[0]

    def get_goniometer(self, index=None):
        return Format.get_goniometer(self)

    def get_detector(self, index=None):
        return Format.get_detector(self)

    def get_beam(self, index=None):
        return Format.get_beam(self)

    def get_scan(self, index=None):
        if index is None:
            return Format.get_scan(self)
        else:
            scan = Format.get_scan(self)
            return scan[index]

    def get_raw_data(self, index):
        detector = self._h5_handle["entry/instrument/detector"]
        data = detector["data"]
        im = data[index, :, :].astype("int32")  # convert from int16
        return flex.int(im)

    def get_image_file(self, index=None):
        return Format.get_image_file(self)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatHDF5Lambda.understand(arg))
